# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import re
from pathlib import Path
from typing import Callable, List, Optional

import pytest

from playwright._impl._glob import glob_to_regex_pattern
from playwright._impl._helper import url_matches
from playwright.async_api import (
    Browser,
    BrowserContext,
    Error,
    Page,
    Playwright,
    Request,
    Route,
    expect,
)
from tests.server import Server, TestServerRequest
from tests.utils import must


async def test_page_route_should_intercept(page: Page, server: Server) -> None:
    intercepted = []

    async def handle_request(route: Route, request: Request) -> None:
        assert route.request == request
        assert "empty.html" in request.url
        assert request.headers["user-agent"]
        assert request.method == "GET"
        assert request.post_data is None
        assert request.is_navigation_request()
        assert request.resource_type == "document"
        assert request.frame == page.main_frame
        assert request.frame.url == "about:blank"
        await route.continue_()
        intercepted.append(True)

    await page.route("**/empty.html", handle_request)

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.ok
    assert len(intercepted) == 1


async def test_page_route_should_unroute(page: Page, server: Server) -> None:
    intercepted = []

    def _handle1(route: Route) -> None:
        intercepted.append(1)
        asyncio.create_task(route.continue_())

    await page.route("**/*", _handle1)

    def _handle2(route: Route, request: Request) -> None:
        intercepted.append(2)
        asyncio.create_task(route.continue_())

    await page.route("**/empty.html", _handle2)

    def _handle3(route: Route, request: Request) -> None:
        intercepted.append(3)
        asyncio.create_task(route.continue_())

    await page.route(
        "**/empty.html",
        _handle3,
    )

    def handler4(route: Route) -> None:
        intercepted.append(4)
        asyncio.create_task(route.continue_())

    await page.route(re.compile("empty.html"), handler4)

    await page.goto(server.EMPTY_PAGE)
    assert intercepted == [4]

    intercepted = []
    await page.unroute(re.compile("empty.html"), handler4)
    await page.goto(server.EMPTY_PAGE)
    assert intercepted == [3]

    intercepted = []
    await page.unroute("**/empty.html")
    await page.goto(server.EMPTY_PAGE)
    assert intercepted == [1]


async def test_page_route_should_work_when_POST_is_redirected_with_302(
    page: Page, server: Server
) -> None:
    server.set_redirect("/rredirect", "/empty.html")
    await page.goto(server.EMPTY_PAGE)
    await page.route("**/*", lambda route: route.continue_())
    await page.set_content(
        """
      <form action='/rredirect' method='post'>
        <input type="hidden" id="foo" name="foo" value="FOOBAR">
      </form>
    """
    )
    async with page.expect_navigation():
        await page.eval_on_selector("form", "form => form.submit()")


# @see https://github.com/GoogleChrome/puppeteer/issues/3973
async def test_page_route_should_work_when_header_manipulation_headers_with_redirect(
    page: Page, server: Server
) -> None:
    server.set_redirect("/rrredirect", "/empty.html")
    await page.route(
        "**/*",
        lambda route: route.continue_(headers={**route.request.headers, "foo": "bar"}),
    )

    await page.goto(server.PREFIX + "/rrredirect")


# @see https://github.com/GoogleChrome/puppeteer/issues/4743
async def test_page_route_should_be_able_to_remove_headers(
    page: Page, server: Server
) -> None:
    async def handle_request(route: Route) -> None:
        headers = route.request.headers
        if "origin" in headers:
            del headers["origin"]
        await route.continue_(headers=headers)

    await page.route(
        "**/*",  # remove "origin" header
        handle_request,
    )

    [serverRequest, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"), page.goto(server.PREFIX + "/empty.html")
    )
    assert serverRequest.getHeader("origin") is None


async def test_page_route_should_contain_referer_header(
    page: Page, server: Server
) -> None:
    requests = []

    def _handle(route: Route, request: Request) -> None:
        requests.append(route.request)
        asyncio.create_task(route.continue_())

    await page.route(
        "**/*",
        _handle,
    )

    await page.goto(server.PREFIX + "/one-style.html")
    assert "/one-style.css" in requests[1].url
    assert "/one-style.html" in requests[1].headers["referer"]


async def test_page_route_should_properly_return_navigation_response_when_URL_has_cookies(
    context: BrowserContext, page: Page, server: Server
) -> None:
    # Setup cookie.
    await page.goto(server.EMPTY_PAGE)
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "foo", "value": "bar"}]
    )

    # Setup request interception.
    await page.route("**/*", lambda route: route.continue_())
    response = await page.reload()
    assert response
    assert response.status == 200


async def test_page_route_should_show_custom_HTTP_headers(
    page: Page, server: Server
) -> None:
    await page.set_extra_http_headers({"foo": "bar"})

    def assert_headers(request: Request) -> None:
        assert request.headers["foo"] == "bar"

    def _handle(route: Route) -> None:
        assert_headers(route.request)
        asyncio.create_task(route.continue_())

    await page.route(
        "**/*",
        _handle,
    )

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.ok


# @see https://github.com/GoogleChrome/puppeteer/issues/4337
async def test_page_route_should_work_with_redirect_inside_sync_XHR(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    server.set_redirect("/logo.png", "/pptr.png")
    await page.route("**/*", lambda route: route.continue_())
    status = await page.evaluate(
        """async() => {
      const request = new XMLHttpRequest();
      request.open('GET', '/logo.png', false);  // `false` makes the request synchronous
      request.send(null);
      return request.status;
    }"""
    )

    assert status == 200


async def test_page_route_should_work_with_custom_referer_headers(
    page: Page, server: Server
) -> None:
    await page.set_extra_http_headers({"referer": server.EMPTY_PAGE})

    def assert_headers(route: Route) -> None:
        assert route.request.headers["referer"] == server.EMPTY_PAGE

    def _handle(route: Route, request: Request) -> None:
        assert_headers(route)
        asyncio.create_task(route.continue_())

    await page.route(
        "**/*",
        _handle,
    )

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.ok


async def test_page_route_should_be_abortable(page: Page, server: Server) -> None:
    await page.route(r"/\.css$/", lambda route: asyncio.create_task(route.abort()))
    failed = []

    def handle_request(request: Request) -> None:
        if ".css" in request.url:
            failed.append(True)

    page.on("requestfailed", handle_request)

    response = await page.goto(server.PREFIX + "/one-style.html")
    assert response
    assert response.ok
    assert response.request.failure is None
    assert len(failed) == 0


async def test_page_route_should_be_abortable_with_custom_error_codes(
    page: Page, server: Server, is_webkit: bool, is_firefox: bool
) -> None:
    await page.route(
        "**/*",
        lambda route: route.abort("internetdisconnected"),
    )
    failed_requests = []
    page.on("requestfailed", lambda request: failed_requests.append(request))
    with pytest.raises(Error):
        await page.goto(server.EMPTY_PAGE)
    assert len(failed_requests) == 1
    failed_request = failed_requests[0]
    if is_webkit:
        assert failed_request.failure == "Blocked by Web Inspector"
    elif is_firefox:
        assert failed_request.failure == "NS_ERROR_OFFLINE"
    else:
        assert failed_request.failure == "net::ERR_INTERNET_DISCONNECTED"


async def test_page_route_should_send_referer(page: Page, server: Server) -> None:
    await page.set_extra_http_headers({"referer": "http://google.com/"})

    await page.route("**/*", lambda route: route.continue_())
    [request, _] = await asyncio.gather(
        server.wait_for_request("/grid.html"),
        page.goto(server.PREFIX + "/grid.html"),
    )
    assert request.getHeader("referer") == "http://google.com/"


async def test_page_route_should_fail_navigation_when_aborting_main_resource(
    page: Page, server: Server, is_webkit: bool, is_firefox: bool
) -> None:
    await page.route("**/*", lambda route: route.abort())
    with pytest.raises(Error) as exc:
        await page.goto(server.EMPTY_PAGE)
    assert exc
    if is_webkit:
        assert "Blocked by Web Inspector" in exc.value.message
    elif is_firefox:
        assert "NS_ERROR_FAILURE" in exc.value.message
    else:
        assert "net::ERR_FAILED" in exc.value.message


async def test_page_route_should_not_work_with_redirects(
    page: Page, server: Server
) -> None:
    intercepted = []

    def _handle(route: Route, request: Request) -> None:
        asyncio.create_task(route.continue_())
        intercepted.append(route.request)

    await page.route(
        "**/*",
        _handle,
    )

    server.set_redirect("/non-existing-page.html", "/non-existing-page-2.html")
    server.set_redirect("/non-existing-page-2.html", "/non-existing-page-3.html")
    server.set_redirect("/non-existing-page-3.html", "/non-existing-page-4.html")
    server.set_redirect("/non-existing-page-4.html", "/empty.html")

    response = await page.goto(server.PREFIX + "/non-existing-page.html")
    assert response
    assert response.status == 200
    assert "empty.html" in response.url

    assert len(intercepted) == 1
    assert intercepted[0].resource_type == "document"
    assert intercepted[0].is_navigation_request()
    assert "/non-existing-page.html" in intercepted[0].url

    chain = []
    r: Optional[Request] = response.request
    while r:
        chain.append(r)
        assert r.is_navigation_request()
        r = r.redirected_from

    assert len(chain) == 5
    assert "/empty.html" in chain[0].url
    assert "/non-existing-page-4.html" in chain[1].url
    assert "/non-existing-page-3.html" in chain[2].url
    assert "/non-existing-page-2.html" in chain[3].url
    assert "/non-existing-page.html" in chain[4].url
    for idx, _ in enumerate(chain):
        assert chain[idx].redirected_to == (chain[idx - 1] if idx > 0 else None)


async def test_page_route_should_work_with_redirects_for_subresources(
    page: Page, server: Server
) -> None:
    intercepted: List[Request] = []

    def _handle(route: Route) -> None:
        asyncio.create_task(route.continue_())
        intercepted.append(route.request)

    await page.route(
        "**/*",
        _handle,
    )

    server.set_redirect("/one-style.css", "/two-style.css")
    server.set_redirect("/two-style.css", "/three-style.css")
    server.set_redirect("/three-style.css", "/four-style.css")
    server.set_route(
        "/four-style.css",
        lambda req: (req.write(b"body {box-sizing: border-box; }"), req.finish()),
    )

    response = await page.goto(server.PREFIX + "/one-style.html")
    assert response
    assert response.status == 200
    assert "one-style.html" in response.url

    # TODO: https://github.com/microsoft/playwright/issues/12789
    assert len(intercepted) >= 2
    assert intercepted[0].resource_type == "document"
    assert "one-style.html" in intercepted[0].url

    r: Optional[Request] = intercepted[1]
    for url in [
        "/one-style.css",
        "/two-style.css",
        "/three-style.css",
        "/four-style.css",
    ]:
        assert r
        assert r.resource_type == "stylesheet"
        assert url in r.url
        r = r.redirected_to
    assert r is None


async def test_page_route_should_work_with_equal_requests(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    hits = [True]

    def handle_request(request: TestServerRequest, hits: List[bool]) -> None:
        request.write(str(len(hits) * 11).encode())
        request.finish()
        hits.append(True)

    server.set_route("/zzz", lambda r: handle_request(r, hits))

    spinner: List[bool] = []

    async def handle_route(route: Route) -> None:
        if len(spinner) == 1:
            await route.abort()
            spinner.pop(0)
        else:
            await route.continue_()
            spinner.append(True)

    # Cancel 2nd request.
    await page.route("**/*", handle_route)

    results = []
    for idx in range(3):
        results.append(
            await page.evaluate(
                """() => fetch('/zzz').then(response => response.text()).catch(e => 'FAILED')"""
            )
        )
    assert results == ["11", "FAILED", "22"]


async def test_page_route_should_navigate_to_dataURL_and_not_fire_dataURL_requests(
    page: Page, server: Server
) -> None:
    requests = []

    def _handle(route: Route) -> None:
        requests.append(route.request)
        asyncio.create_task(route.continue_())

    await page.route(
        "**/*",
        _handle,
    )

    data_URL = "data:text/html,<div>yo</div>"
    response = await page.goto(data_URL)
    assert response is None
    assert len(requests) == 0


async def test_page_route_should_be_able_to_fetch_dataURL_and_not_fire_dataURL_requests(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    requests = []

    def _handle(route: Route) -> None:
        requests.append(route.request)
        asyncio.create_task(route.continue_())

    await page.route("**/*", _handle)

    data_URL = "data:text/html,<div>yo</div>"
    text = await page.evaluate("url => fetch(url).then(r => r.text())", data_URL)
    assert text == "<div>yo</div>"
    assert len(requests) == 0


async def test_page_route_should_navigate_to_URL_with_hash_and_and_fire_requests_without_hash(
    page: Page, server: Server
) -> None:
    requests = []

    def _handle(route: Route) -> None:
        requests.append(route.request)
        asyncio.create_task(route.continue_())

    await page.route(
        "**/*",
        _handle,
    )

    response = await page.goto(server.EMPTY_PAGE + "#hash")
    assert response
    assert response.status == 200
    assert response.url == server.EMPTY_PAGE
    assert len(requests) == 1
    assert requests[0].url == server.EMPTY_PAGE


async def test_page_route_should_work_with_encoded_server(
    page: Page, server: Server
) -> None:
    # The requestWillBeSent will report encoded URL, whereas interception will
    # report URL as-is. @see crbug.com/759388
    await page.route("**/*", lambda route: route.continue_())
    response = await page.goto(server.PREFIX + "/some nonexisting page")
    assert response
    assert response.status == 404


async def test_page_route_should_work_with_encoded_server___2(
    page: Page, server: Server
) -> None:
    # The requestWillBeSent will report URL as-is, whereas interception will
    # report encoded URL for stylesheet. @see crbug.com/759388
    requests: List[Request] = []

    def _handle(route: Route) -> None:
        asyncio.create_task(route.continue_())
        requests.append(route.request)

    await page.route("**/*", _handle)

    response = await page.goto(
        f"""data:text/html,<link rel="stylesheet" href="{server.PREFIX}/fonts?helvetica|arial"/>"""
    )
    assert response is None
    # TODO: https://github.com/microsoft/playwright/issues/12789
    assert len(requests) >= 1
    assert (must(await requests[0].response())).status == 404


async def test_page_route_should_not_throw_Invalid_Interception_Id_if_the_request_was_cancelled(
    page: Page, server: Server
) -> None:
    await page.set_content("<iframe></iframe>")
    route_future: "asyncio.Future[Route]" = asyncio.Future()
    await page.route("**/*", lambda r, _: route_future.set_result(r))

    async with page.expect_request("**/*"):
        await page.eval_on_selector(
            "iframe", """(frame, url) => frame.src = url""", server.EMPTY_PAGE
        )
    # Delete frame to cause request to be canceled.
    await page.eval_on_selector("iframe", "frame => frame.remove()")
    route = await route_future
    await route.continue_()


async def test_page_route_should_intercept_main_resource_during_cross_process_navigation(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    intercepted = []

    def _handle(route: Route) -> None:
        intercepted.append(True)
        asyncio.create_task(route.continue_())

    await page.route(
        server.CROSS_PROCESS_PREFIX + "/empty.html",
        _handle,
    )

    response = await page.goto(server.CROSS_PROCESS_PREFIX + "/empty.html")
    assert response
    assert response.ok
    assert len(intercepted) == 1


@pytest.mark.skip_browser("webkit")
async def test_page_route_should_create_a_redirect(page: Page, server: Server) -> None:
    await page.goto(server.PREFIX + "/empty.html")

    async def handle_route(route: Route, request: Request) -> None:
        if request.url != (server.PREFIX + "/redirect_this"):
            return await route.continue_()
        await route.fulfill(status=301, headers={"location": "/empty.html"})

    await page.route(
        "**/*",
        handle_route,
    )

    text = await page.evaluate(
        """async url => {
      const data = await fetch(url);
      return data.text();
    }""",
        server.PREFIX + "/redirect_this",
    )
    assert text == ""


async def test_page_route_should_support_cors_with_GET(
    page: Page, server: Server, browser_name: str
) -> None:
    await page.goto(server.EMPTY_PAGE)

    async def handle_route(route: Route, request: Request) -> None:
        headers = {
            "access-control-allow-origin": (
                "*" if request.url.endswith("allow") else "none"
            )
        }
        await route.fulfill(
            content_type="application/json",
            headers=headers,
            status=200,
            body=json.dumps(["electric", "gas"]),
        )

    await page.route(
        "**/cars*",
        handle_route,
    )
    # Should succeed
    resp = await page.evaluate(
        """async () => {
        const response = await fetch('https://example.com/cars?allow', { mode: 'cors' });
        return response.json();
      }"""
    )

    assert resp == ["electric", "gas"]

    # Should be rejected
    with pytest.raises(Error) as exc:
        await page.evaluate(
            """async () => {
            const response = await fetch('https://example.com/cars?reject', { mode: 'cors' });
            return response.json();
        }"""
        )
    if browser_name == "chromium":
        assert "Failed" in exc.value.message
    elif browser_name == "webkit":
        assert "TypeError" in exc.value.message
    elif browser_name == "firefox":
        assert "NetworkError" in exc.value.message


async def test_page_route_should_support_cors_with_POST(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await page.route(
        "**/cars",
        lambda route: route.fulfill(
            content_type="application/json",
            headers={"Access-Control-Allow-Origin": "*"},
            status=200,
            body=json.dumps(["electric", "gas"]),
        ),
    )

    resp = await page.evaluate(
        """async () => {
      const response = await fetch('https://example.com/cars', {
        method: 'POST',
        headers: { 'Content-Type': 'application/json' },
        mode: 'cors',
        body: JSON.stringify({ 'number': 1 })
      });
      return response.json();
    }"""
    )

    assert resp == ["electric", "gas"]


async def test_page_route_should_support_cors_for_different_methods(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await page.route(
        "**/cars",
        lambda route, request: route.fulfill(
            content_type="application/json",
            headers={"Access-Control-Allow-Origin": "*"},
            status=200,
            body=json.dumps([request.method, "electric", "gas"]),
        ),
    )

    # First POST
    resp = await page.evaluate(
        """async () => {
        const response = await fetch('https://example.com/cars', {
          method: 'POST',
          headers: { 'Content-Type': 'application/json' },
          mode: 'cors',
          body: JSON.stringify({ 'number': 1 })
        });
        return response.json();
      }"""
    )

    assert resp == ["POST", "electric", "gas"]
    # Then DELETE
    resp = await page.evaluate(
        """async () => {
        const response = await fetch('https://example.com/cars', {
          method: 'DELETE',
          headers: {},
          mode: 'cors',
          body: ''
        });
        return response.json();
      }"""
    )

    assert resp == ["DELETE", "electric", "gas"]


async def test_request_fulfill_should_work_a(page: Page, server: Server) -> None:
    await page.route(
        "**/*",
        lambda route: route.fulfill(
            status=201,
            headers={"foo": "bar"},
            content_type="text/html",
            body="Yo, page!",
        ),
    )

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.status == 201
    assert response.headers["foo"] == "bar"
    assert await page.evaluate("() => document.body.textContent") == "Yo, page!"


async def test_request_fulfill_should_work_with_status_code_422(
    page: Page, server: Server
) -> None:
    await page.route(
        "**/*",
        lambda route: route.fulfill(status=422, body="Yo, page!"),
    )

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.status == 422
    assert response.status_text == "Unprocessable Entity"
    assert await page.evaluate("() => document.body.textContent") == "Yo, page!"


async def test_request_fulfill_should_allow_mocking_binary_responses(
    page: Page,
    server: Server,
    assert_to_be_golden: Callable[[bytes, str], None],
    assetdir: Path,
) -> None:
    await page.route(
        "**/*",
        lambda route: route.fulfill(
            content_type="image/png",
            body=(assetdir / "pptr.png").read_bytes(),
        ),
    )

    await page.evaluate(
        """PREFIX => {
      const img = document.createElement('img');
      img.src = PREFIX + '/does-not-exist.png';
      document.body.appendChild(img);
      return new Promise(fulfill => img.onload = fulfill);
    }""",
        server.PREFIX,
    )
    img = await page.query_selector("img")
    assert img
    assert_to_be_golden(await img.screenshot(), "mock-binary-response.png")


async def test_request_fulfill_should_allow_mocking_svg_with_charset(
    page: Page, server: Server, assert_to_be_golden: Callable[[bytes, str], None]
) -> None:
    await page.route(
        "**/*",
        lambda route: route.fulfill(
            content_type="image/svg+xml ; charset=utf-8",
            body='<svg width="50" height="50" version="1.1" xmlns="http://www.w3.org/2000/svg"><rect x="10" y="10" width="30" height="30" stroke="black" fill="transparent" stroke-width="5"/></svg>',
        ),
    )

    await page.evaluate(
        """PREFIX => {
      const img = document.createElement('img');
      img.src = PREFIX + '/does-not-exist.svg';
      document.body.appendChild(img);
      return new Promise((f, r) => { img.onload = f; img.onerror = r; });
    }""",
        server.PREFIX,
    )
    img = await page.query_selector("img")
    assert img
    assert_to_be_golden(await img.screenshot(), "mock-svg.png")


async def test_request_fulfill_should_work_with_file_path(
    page: Page,
    server: Server,
    assert_to_be_golden: Callable[[bytes, str], None],
    assetdir: Path,
) -> None:
    await page.route(
        "**/*",
        lambda route: route.fulfill(
            content_type="shouldBeIgnored", path=assetdir / "pptr.png"
        ),
    )
    await page.evaluate(
        """PREFIX => {
      const img = document.createElement('img');
      img.src = PREFIX + '/does-not-exist.png';
      document.body.appendChild(img);
      return new Promise(fulfill => img.onload = fulfill);
    }""",
        server.PREFIX,
    )
    img = await page.query_selector("img")
    assert img
    assert_to_be_golden(await img.screenshot(), "mock-binary-response.png")


async def test_request_fulfill_should_stringify_intercepted_request_response_headers(
    page: Page, server: Server
) -> None:
    await page.route(
        "**/*",
        lambda route: route.fulfill(
            status=200, headers={"foo": True}, body="Yo, page!"  # type: ignore
        ),
    )

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.status == 200
    headers = response.headers
    assert headers["foo"] == "True"
    assert await page.evaluate("() => document.body.textContent") == "Yo, page!"


async def test_request_fulfill_should_not_modify_the_headers_sent_to_the_server(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/empty.html")
    interceptedRequests = []

    # this is just to enable request interception, which disables caching in chromium
    await page.route(server.PREFIX + "/unused", lambda route, req: None)

    def _handler1(response: TestServerRequest) -> None:
        interceptedRequests.append(response)
        response.setHeader("Access-Control-Allow-Origin", "*")
        response.write(b"done")
        response.finish()

    server.set_route("/something", _handler1)

    text = await page.evaluate(
        """async url => {
      const data = await fetch(url);
      return data.text();
    }""",
        server.CROSS_PROCESS_PREFIX + "/something",
    )
    assert text == "done"

    playwrightRequest: "asyncio.Future[Request]" = asyncio.Future()

    def _handler2(route: Route, request: Request) -> None:
        playwrightRequest.set_result(request)
        asyncio.create_task(route.continue_(headers={**request.headers}))

    await page.route(
        server.CROSS_PROCESS_PREFIX + "/something",
        _handler2,
    )

    textAfterRoute = await page.evaluate(
        """async url => {
      const data = await fetch(url);
      return data.text();
    }""",
        server.CROSS_PROCESS_PREFIX + "/something",
    )
    assert textAfterRoute == "done"

    assert len(interceptedRequests) == 2
    assert (
        interceptedRequests[0].requestHeaders == interceptedRequests[1].requestHeaders
    )


async def test_request_fulfill_should_include_the_origin_header(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/empty.html")
    interceptedRequest = []

    def _handle(route: Route, request: Request) -> None:
        interceptedRequest.append(request)
        asyncio.create_task(
            route.fulfill(
                headers={"Access-Control-Allow-Origin": "*"},
                content_type="text/plain",
                body="done",
            )
        )

    await page.route(server.CROSS_PROCESS_PREFIX + "/something", _handle)

    text = await page.evaluate(
        """async url => {
      const data = await fetch(url);
      return data.text();
    }""",
        server.CROSS_PROCESS_PREFIX + "/something",
    )
    assert text == "done"
    assert len(interceptedRequest) == 1
    assert interceptedRequest[0].headers["origin"] == server.PREFIX


async def test_request_fulfill_should_work_with_request_interception(
    page: Page, server: Server
) -> None:
    requests = {}

    async def _handle_route(route: Route) -> None:
        requests[route.request.url.split("/").pop()] = route.request
        await route.continue_()

    await page.route("**/*", _handle_route)

    server.set_redirect("/rrredirect", "/frames/one-frame.html")
    await page.goto(server.PREFIX + "/rrredirect")
    assert requests["rrredirect"].is_navigation_request()
    assert requests["frame.html"].is_navigation_request()
    assert requests["script.js"].is_navigation_request() is False
    assert requests["style.css"].is_navigation_request() is False


async def test_Interception_should_work_with_request_interception(
    browser: Browser, https_server: Server
) -> None:
    context = await browser.new_context(ignore_https_errors=True)
    page = await context.new_page()

    await page.route("**/*", lambda route: asyncio.ensure_future(route.continue_()))
    response = await page.goto(https_server.EMPTY_PAGE)
    assert response
    assert response.status == 200
    await context.close()


async def test_ignore_http_errors_service_worker_should_intercept_after_a_service_worker(
    page: Page, server: Server
) -> None:
    await page.goto(server.PREFIX + "/serviceworkers/fetchdummy/sw.html")
    await page.evaluate("() => window.activationPromise")

    # Sanity check.
    sw_response = await page.evaluate('() => fetchDummy("foo")')
    assert sw_response == "responseFromServiceWorker:foo"

    def _handle_route(route: Route) -> None:
        asyncio.ensure_future(
            route.fulfill(
                status=200,
                content_type="text/css",
                body="responseFromInterception:" + route.request.url.split("/")[-1],
            )
        )

    await page.route("**/foo", _handle_route)

    # Page route is applied after service worker fetch event.
    sw_response2 = await page.evaluate('() => fetchDummy("foo")')
    assert sw_response2 == "responseFromServiceWorker:foo"

    # Page route is not applied to service worker initiated fetch.
    non_intercepted_response = await page.evaluate('() => fetchDummy("passthrough")')
    assert non_intercepted_response == "FAILURE: Not Found"


async def test_page_route_should_support_times_parameter(
    page: Page, server: Server
) -> None:
    intercepted = []

    async def handle_request(route: Route) -> None:
        await route.continue_()
        intercepted.append(True)

    await page.route("**/empty.html", handle_request, times=1)

    await page.goto(server.EMPTY_PAGE)
    await page.goto(server.EMPTY_PAGE)
    await page.goto(server.EMPTY_PAGE)
    assert len(intercepted) == 1


async def test_should_work_if_handler_with_times_parameter_was_removed_from_another_handler(
    context: BrowserContext, page: Page, server: Server
) -> None:
    intercepted = []

    async def handler(route: Route) -> None:
        intercepted.append("first")
        await route.continue_()

    await page.route("**/*", handler, times=1)

    async def handler2(route: Route) -> None:
        intercepted.append("second")
        await page.unroute("**/*", handler)
        await route.fallback()

    await page.route("**/*", handler2)
    await page.goto(server.EMPTY_PAGE)
    assert intercepted == ["second"]
    intercepted.clear()
    await page.goto(server.EMPTY_PAGE)
    assert intercepted == ["second"]


async def test_should_fulfill_with_global_fetch_result(
    page: Page, playwright: Playwright, server: Server
) -> None:
    async def handle_request(route: Route) -> None:
        request = await playwright.request.new_context()
        response = await request.get(server.PREFIX + "/simple.json")
        await route.fulfill(response=response)
        await request.dispose()

    await page.route("**/*", handle_request)

    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.status == 200
    assert await response.json() == {"foo": "bar"}


async def test_should_work_with_glob() -> None:
    def glob_to_regex(pattern: str) -> re.Pattern:
        return re.compile(glob_to_regex_pattern(pattern))

    assert glob_to_regex("**/*.js").match("https://localhost:8080/foo.js")
    assert not glob_to_regex("**/*.css").match("https://localhost:8080/foo.js")
    assert not glob_to_regex("*.js").match(
        "https://localhost:8080/foo.js"
    )  # Doesn"t match path separator
    assert glob_to_regex("https://**/*.js").match("https://localhost:8080/foo.js")
    assert glob_to_regex("http://localhost:8080/simple/path.js").match(
        "http://localhost:8080/simple/path.js"
    )
    assert glob_to_regex("**/{a,b}.js").match("https://localhost:8080/a.js")
    assert glob_to_regex("**/{a,b}.js").match("https://localhost:8080/b.js")
    assert not glob_to_regex("**/{a,b}.js").match("https://localhost:8080/c.js")

    assert glob_to_regex("**/*.{png,jpg,jpeg}").match("https://localhost:8080/c.jpg")
    assert glob_to_regex("**/*.{png,jpg,jpeg}").match("https://localhost:8080/c.jpeg")
    assert glob_to_regex("**/*.{png,jpg,jpeg}").match("https://localhost:8080/c.png")
    assert not glob_to_regex("**/*.{png,jpg,jpeg}").match(
        "https://localhost:8080/c.css"
    )
    assert glob_to_regex("foo*").match("foo.js")
    assert not glob_to_regex("foo*").match("foo/bar.js")
    assert not glob_to_regex("http://localhost:3000/signin-oidc*").match(
        "http://localhost:3000/signin-oidc/foo"
    )
    assert glob_to_regex("http://localhost:3000/signin-oidc*").match(
        "http://localhost:3000/signin-oidcnice"
    )

    # range [] is NOT supported
    assert glob_to_regex("**/api/v[0-9]").fullmatch("http://example.com/api/v[0-9]")
    assert not glob_to_regex("**/api/v[0-9]").fullmatch(
        "http://example.com/api/version"
    )
    assert not glob_to_regex("**/api/v[0-9]").fullmatch(
        "http://example.com/api/v1"
    )  # Should not match if [] is literal

    # query params
    assert glob_to_regex("**/api\\?param").match("http://example.com/api?param")
    assert not glob_to_regex("**/api\\?param").match("http://example.com/api-param")

    assert glob_to_regex("**/three-columns/settings.html\\?**id=settings-**").match(
        "http://mydomain:8080/blah/blah/three-columns/settings.html?id=settings-e3c58efe-02e9-44b0-97ac-dd138100cf7c&blah"
    )

    assert glob_to_regex("\\?").pattern == r"^\?$"
    assert glob_to_regex("\\").pattern == r"^\\$"
    assert glob_to_regex("\\\\").pattern == r"^\\$"
    assert glob_to_regex("\\[").pattern == r"^\[$"
    assert glob_to_regex("[a-z]").pattern == r"^\[a-z\]$"
    assert (
        glob_to_regex("$^+.\\*()|\\?\\{\\}\\[\\]").pattern
        == r"^\$\^\+\.\*\(\)\|\?\{\}\[\]$"
    )

    # --- url_matches tests ---
    # Basic exact and wildcard matching
    assert url_matches(None, "http://playwright.dev/", "http://playwright.dev")
    assert url_matches(None, "http://playwright.dev/?a=b", "http://playwright.dev?a=b")
    assert url_matches(None, "http://playwright.dev/", "h*://playwright.dev")
    assert url_matches(
        None, "http://api.playwright.dev/?x=y", "http://*.playwright.dev?x=y"
    )
    assert url_matches(None, "http://playwright.dev/foo/bar", "**/foo/**")

    # Relative path matching with base URL
    assert url_matches("http://playwright.dev", "http://playwright.dev/?x=y", "?x=y")
    assert url_matches(
        "http://playwright.dev/foo/", "http://playwright.dev/foo/bar?x=y", "./bar?x=y"
    )

    # Case insensitive matching
    assert url_matches(
        None, "https://playwright.dev/fooBAR", "HtTpS://pLaYwRiGhT.dEv/fooBAR"
    )
    assert url_matches(
        "http://ignored",
        "https://playwright.dev/fooBAR",
        "HtTpS://pLaYwRiGhT.dEv/fooBAR",
    )
    # Path and search query are case-sensitive
    assert not url_matches(
        None, "https://playwright.dev/foobar", "https://playwright.dev/fooBAR"
    )
    assert not url_matches(
        None, "https://playwright.dev/foobar?a=b", "https://playwright.dev/foobar?A=B"
    )

    # This is not supported, we treat ? as a query separator.
    assert not url_matches(
        None,
        "http://localhost:8080/Simple/path.js",
        "http://localhost:8080/?imple/path.js",
    )
    assert not url_matches(None, "http://playwright.dev/", "http://playwright.?ev")
    assert url_matches(None, "http://playwright./?ev", "http://playwright.?ev")
    assert not url_matches(
        None, "http://playwright.dev/foo", "http://playwright.dev/f??"
    )
    assert url_matches(None, "http://playwright.dev/f??", "http://playwright.dev/f??")
    assert url_matches(
        None, "http://playwright.dev/?x=y", r"http://playwright.dev\?x=y"
    )
    assert url_matches(
        None, "http://playwright.dev/?x=y", r"http://playwright.dev/\?x=y"
    )
    assert url_matches(
        "http://playwright.dev/foo", "http://playwright.dev/foo?bar", "?bar"
    )
    assert url_matches(
        "http://playwright.dev/foo", "http://playwright.dev/foo?bar", r"\\?bar"
    )
    assert url_matches("http://first.host/", "http://second.host/foo", "**/foo")
    assert url_matches("http://playwright.dev/", "http://localhost/", "*//localhost/")

    custom_prefixes = ["about", "data", "chrome", "edge", "file"]
    for prefix in custom_prefixes:
        assert url_matches(
            "http://playwright.dev/", f"{prefix}:blank", f"{prefix}:blank"
        )
        assert not url_matches(
            "http://playwright.dev/", f"{prefix}:blank", "http://playwright.dev/"
        )
        assert url_matches(None, f"{prefix}:blank", f"{prefix}:blank")
        assert url_matches(None, f"{prefix}:blank", f"{prefix}:*")
        assert not url_matches(None, f"not{prefix}:blank", f"{prefix}:*")

    # Added for Python implementation
    assert url_matches(
        None,
        "custom://example.com/foo/bar?id=123",
        "{custom,another}://example.com/foo/bar?id=123",
    )
    assert not url_matches(
        None, "custom://example.com/foo/bar?id=123", "**example.com/foo/bar?id=123"
    )


async def test_should_not_support_question_in_glob_pattern(
    page: Page, playwright: Playwright, server: Server
) -> None:
    server.set_route("/index", lambda req: (req.write(b"index-no-hello"), req.finish()))
    server.set_route(
        "/index123hello", lambda req: (req.write(b"index123hello"), req.finish())
    )
    server.set_route(
        "/index?hello", lambda req: (req.write(b"index?hello"), req.finish())
    )
    server.set_route(
        "/index1hello", lambda req: (req.write(b"index1hello"), req.finish())
    )

    async def handle_any_char(route: Route) -> None:
        await route.fulfill(body="intercepted any character")

    await page.route("**/index?hello", handle_any_char)

    async def handle_question_mark(route: Route) -> None:
        await route.fulfill(body="intercepted question mark")

    await page.route(r"**/index\?hello", handle_question_mark)

    await page.goto(server.PREFIX + "/index?hello")
    await expect(page.locator("body")).to_have_text("intercepted question mark")

    await page.goto(server.PREFIX + "/index")
    await expect(page.locator("body")).to_have_text("index-no-hello")

    await page.goto(server.PREFIX + "/index1hello")
    await expect(page.locator("body")).to_have_text("index1hello")

    await page.goto(server.PREFIX + "/index123hello")
    await expect(page.locator("body")).to_have_text("index123hello")



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from playwright.async_api import Page, TimeoutError, expect
from tests.server import Server


async def test_to_be_in_viewport_should_work(page: Page, server: Server) -> None:
    await page.set_content(
        """
      <div id=big style="height: 10000px;"></div>
      <div id=small>foo</div>
    """
    )
    await expect(page.locator("#big")).to_be_in_viewport()
    await expect(page.locator("#small")).not_to_be_in_viewport()
    await page.locator("#small").scroll_into_view_if_needed()
    await expect(page.locator("#small")).to_be_in_viewport()
    await expect(page.locator("#small")).to_be_in_viewport(ratio=1)


async def test_to_be_in_viewport_should_respect_ratio_option(
    page: Page, server: Server
) -> None:
    await page.set_content(
        """
      <style>body, div, html { padding: 0; margin: 0; }</style>
      <div id=big style="height: 400vh;"></div>
    """
    )
    await expect(page.locator("div")).to_be_in_viewport()
    await expect(page.locator("div")).to_be_in_viewport(ratio=0.1)
    await expect(page.locator("div")).to_be_in_viewport(ratio=0.2)

    await expect(page.locator("div")).to_be_in_viewport(ratio=0.25)
    # In this test, element's ratio is 0.25.
    await expect(page.locator("div")).not_to_be_in_viewport(ratio=0.26)

    await expect(page.locator("div")).not_to_be_in_viewport(ratio=0.3)
    await expect(page.locator("div")).not_to_be_in_viewport(ratio=0.7)
    await expect(page.locator("div")).not_to_be_in_viewport(ratio=0.8)


async def test_to_be_in_viewport_should_have_good_stack(
    page: Page, server: Server
) -> None:
    with pytest.raises(AssertionError) as exc_info:
        await expect(page.locator("body")).not_to_be_in_viewport(timeout=100)
    assert 'unexpected value "viewport ratio' in str(exc_info.value)


async def test_to_be_in_viewport_should_report_intersection_even_if_fully_covered_by_other_element(
    page: Page, server: Server
) -> None:
    await page.set_content(
        """
      <h1>hello</h1>
      <div style="position: relative; height: 10000px; top: -5000px;></div>
    """
    )
    await expect(page.locator("h1")).to_be_in_viewport()


async def test_should_have_timeout_error_name(page: Page) -> None:
    with pytest.raises(TimeoutError) as exc_info:
        await page.wait_for_selector("#not-found", timeout=1)
    assert exc_info.value.name == "TimeoutError"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from urllib.parse import urlparse

from playwright.async_api import Browser, BrowserContext, Page
from tests.server import Server


async def test_should_clear_cookies(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "cookie1", "value": "1"}]
    )
    assert await page.evaluate("document.cookie") == "cookie1=1"
    await context.clear_cookies()
    assert await context.cookies() == []
    await page.reload()
    assert await page.evaluate("document.cookie") == ""


async def test_should_isolate_cookies_when_clearing(
    context: BrowserContext, server: Server, browser: Browser
) -> None:
    another_context = await browser.new_context()
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "page1cookie", "value": "page1value"}]
    )
    await another_context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "page2cookie", "value": "page2value"}]
    )

    assert len(await context.cookies()) == 1
    assert len(await another_context.cookies()) == 1

    await context.clear_cookies()
    assert len(await context.cookies()) == 0
    assert len(await another_context.cookies()) == 1

    await another_context.clear_cookies()
    assert len(await context.cookies()) == 0
    assert len(await another_context.cookies()) == 0
    await another_context.close()


async def test_should_remove_cookies_by_name(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await context.add_cookies(
        [
            {
                "name": "cookie1",
                "value": "1",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/",
            },
            {
                "name": "cookie2",
                "value": "2",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/",
            },
        ]
    )
    await page.goto(server.PREFIX)
    assert await page.evaluate("document.cookie") == "cookie1=1; cookie2=2"
    await context.clear_cookies(name="cookie1")
    assert await page.evaluate("document.cookie") == "cookie2=2"


async def test_should_remove_cookies_by_name_regex(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await context.add_cookies(
        [
            {
                "name": "cookie1",
                "value": "1",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/",
            },
            {
                "name": "cookie2",
                "value": "2",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/",
            },
        ]
    )
    await page.goto(server.PREFIX)
    assert await page.evaluate("document.cookie") == "cookie1=1; cookie2=2"
    await context.clear_cookies(name=re.compile("coo.*1"))
    assert await page.evaluate("document.cookie") == "cookie2=2"


async def test_should_remove_cookies_by_domain(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await context.add_cookies(
        [
            {
                "name": "cookie1",
                "value": "1",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/",
            },
            {
                "name": "cookie2",
                "value": "2",
                "domain": urlparse(server.CROSS_PROCESS_PREFIX).hostname,
                "path": "/",
            },
        ]
    )
    await page.goto(server.PREFIX)
    assert await page.evaluate("document.cookie") == "cookie1=1"
    await page.goto(server.CROSS_PROCESS_PREFIX)
    assert await page.evaluate("document.cookie") == "cookie2=2"
    await context.clear_cookies(domain=urlparse(server.CROSS_PROCESS_PREFIX).hostname)
    assert await page.evaluate("document.cookie") == ""
    await page.goto(server.PREFIX)
    assert await page.evaluate("document.cookie") == "cookie1=1"


async def test_should_remove_cookies_by_path(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await context.add_cookies(
        [
            {
                "name": "cookie1",
                "value": "1",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/api/v1",
            },
            {
                "name": "cookie2",
                "value": "2",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/api/v2",
            },
            {
                "name": "cookie3",
                "value": "3",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/",
            },
        ]
    )
    await page.goto(server.PREFIX + "/api/v1")
    assert await page.evaluate("document.cookie") == "cookie1=1; cookie3=3"
    await context.clear_cookies(path="/api/v1")
    assert await page.evaluate("document.cookie") == "cookie3=3"
    await page.goto(server.PREFIX + "/api/v2")
    assert await page.evaluate("document.cookie") == "cookie2=2; cookie3=3"
    await page.goto(server.PREFIX + "/")
    assert await page.evaluate("document.cookie") == "cookie3=3"


async def test_should_remove_cookies_by_name_and_domain(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await context.add_cookies(
        [
            {
                "name": "cookie1",
                "value": "1",
                "domain": urlparse(server.PREFIX).hostname,
                "path": "/",
            },
            {
                "name": "cookie1",
                "value": "1",
                "domain": urlparse(server.CROSS_PROCESS_PREFIX).hostname,
                "path": "/",
            },
        ]
    )
    await page.goto(server.PREFIX)
    assert await page.evaluate("document.cookie") == "cookie1=1"
    await context.clear_cookies(name="cookie1", domain=urlparse(server.PREFIX).hostname)
    assert await page.evaluate("document.cookie") == ""
    await page.goto(server.CROSS_PROCESS_PREFIX)
    assert await page.evaluate("document.cookie") == "cookie1=1"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import gc
import sys
from typing import Dict

import pytest

from playwright.async_api import Page, async_playwright
from tests.server import Server
from tests.utils import TARGET_CLOSED_ERROR_MESSAGE


async def test_should_cancel_underlying_protocol_calls(
    browser_name: str, launch_arguments: Dict
) -> None:
    handler_exception = None

    def exception_handlerdler(loop: asyncio.AbstractEventLoop, context: Dict) -> None:
        nonlocal handler_exception
        handler_exception = context["exception"]

    asyncio.get_running_loop().set_exception_handler(exception_handlerdler)

    async with async_playwright() as p:
        browser = await p[browser_name].launch(**launch_arguments)
        page = await browser.new_page()
        task = asyncio.create_task(page.wait_for_selector("will-never-find"))
        # make sure that the wait_for_selector message was sent to the server (driver)
        await asyncio.sleep(0.1)
        task.cancel()
        with pytest.raises(asyncio.CancelledError):
            await task
        await browser.close()

    # The actual 'Future exception was never retrieved' is logged inside the Future destructor (__del__).
    gc.collect()

    assert handler_exception is None

    asyncio.get_running_loop().set_exception_handler(None)


async def test_async_playwright_stop_multiple_times() -> None:
    playwright = await async_playwright().start()
    await playwright.stop()
    await playwright.stop()


async def test_cancel_pending_protocol_call_on_playwright_stop(server: Server) -> None:
    server.set_route("/hang", lambda _: None)
    playwright = await async_playwright().start()
    api_request_context = await playwright.request.new_context()
    pending_task = asyncio.create_task(api_request_context.get(server.PREFIX + "/hang"))
    await playwright.stop()
    with pytest.raises(Exception) as exc_info:
        await pending_task
    assert TARGET_CLOSED_ERROR_MESSAGE in str(exc_info.value)


async def test_should_not_throw_with_taskgroup(page: Page) -> None:
    if sys.version_info < (3, 11):
        pytest.skip("TaskGroup is only available in Python 3.11+")

    from builtins import ExceptionGroup  # type: ignore

    async def raise_exception() -> None:
        raise ValueError("Something went wrong")

    with pytest.raises(ExceptionGroup) as exc_info:
        async with asyncio.TaskGroup() as group:  # type: ignore
            group.create_task(page.locator(".this-element-does-not-exist").inner_text())
            group.create_task(raise_exception())
    assert len(exc_info.value.exceptions) == 1
    assert "Something went wrong" in str(exc_info.value.exceptions[0])
    assert isinstance(exc_info.value.exceptions[0], ValueError)
    assert await page.evaluate("() => 11 * 11") == 121


async def test_should_return_proper_api_name_on_error(page: Page) -> None:
    try:
        await page.evaluate("does_not_exist")

        assert (
            False
        ), "Accessing undefined JavaScript variable should have thrown exception"
    except Exception as error:
        # Each browser returns slightly different error messages, but they should all start with "Page.evaluate:", because that was the Playwright method where the error originated
        assert str(error).startswith("Page.evaluate:")



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

import pytest

from playwright.async_api import ConsoleMessage, Page
from tests.server import Server


async def test_console_should_work(page: Page, browser_name: str) -> None:
    messages: List[ConsoleMessage] = []
    page.once("console", lambda m: messages.append(m))
    async with page.expect_console_message() as message_info:
        await page.evaluate('() => console.log("hello", 5, {foo: "bar"})')
    message = await message_info.value
    if browser_name != "firefox":
        assert message.text == "hello 5 {foo: bar}"
        assert str(message) == "hello 5 {foo: bar}"
    else:
        assert message.text == "hello 5 JSHandle@object"
        assert str(message) == "hello 5 JSHandle@object"
    assert message.type == "log"
    assert await message.args[0].json_value() == "hello"
    assert await message.args[1].json_value() == 5
    assert await message.args[2].json_value() == {"foo": "bar"}


async def test_console_should_emit_same_log_twice(page: Page) -> None:
    messages = []
    page.on("console", lambda m: messages.append(m.text))
    await page.evaluate('() => { for (let i = 0; i < 2; ++i ) console.log("hello"); } ')
    assert messages == ["hello", "hello"]


async def test_console_should_use_text_for__str__(page: Page) -> None:
    messages = []
    page.on("console", lambda m: messages.append(m))
    await page.evaluate('() => console.log("Hello world")')
    assert len(messages) == 1
    assert str(messages[0]) == "Hello world"


async def test_console_should_work_for_different_console_api_calls(page: Page) -> None:
    messages: List[ConsoleMessage] = []
    page.on("console", lambda m: messages.append(m))
    # All console events will be reported before 'page.evaluate' is finished.
    await page.evaluate(
        """() => {
      // A pair of time/timeEnd generates only one Console API call.
      console.time('calling console.time');
      console.timeEnd('calling console.time');
      console.trace('calling console.trace');
      console.dir('calling console.dir');
      console.warn('calling console.warn');
      console.error('calling console.error');
      console.log(Promise.resolve('should not wait until resolved!'));
    }"""
    )
    assert list(map(lambda msg: msg.type, messages)) == [
        "timeEnd",
        "trace",
        "dir",
        "warning",
        "error",
        "log",
    ]

    assert "calling console.time" in messages[0].text
    assert list(map(lambda msg: msg.text, messages[1:])) == [
        "calling console.trace",
        "calling console.dir",
        "calling console.warn",
        "calling console.error",
        "Promise",
    ]


async def test_console_should_not_fail_for_window_object(
    page: Page, browser_name: str
) -> None:
    async with page.expect_console_message() as message_info:
        await page.evaluate("console.error(window)")
    message = await message_info.value
    if browser_name != "firefox":
        assert message.text == "Window"
    else:
        assert message.text == "JSHandle@object"


# Upstream issue https://bugs.webkit.org/show_bug.cgi?id=229515
@pytest.mark.skip_browser("webkit")
async def test_console_should_trigger_correct_log(page: Page, server: Server) -> None:
    await page.goto("about:blank")
    async with page.expect_console_message() as message_info:
        await page.evaluate("async url => fetch(url).catch(e => {})", server.EMPTY_PAGE)
    message = await message_info.value
    assert "Access-Control-Allow-Origin" in message.text
    assert message.type == "error"


async def test_console_should_have_location_for_console_api_calls(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_console_message() as message_info:
        await page.goto(server.PREFIX + "/consolelog.html")
    message = await message_info.value
    assert message.text == "yellow"
    assert message.type == "log"
    location = message.location
    # Engines have different column notion.
    assert location["url"] == server.PREFIX + "/consolelog.html"
    assert location["lineNumber"] == 7


async def test_console_should_not_throw_when_there_are_console_messages_in_detached_iframes(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_popup() as page_info:
        await page.evaluate(
            """async() => {
                // 1. Create a popup that Playwright is not connected to.
                const win = window.open('');
                window._popup = win;
                if (window.document.readyState !== 'complete')
                await new Promise(f => window.addEventListener('load', f));
                // 2. In this popup, create an iframe that console.logs a message.
                win.document.body.innerHTML = `<iframe src='/consolelog.html'></iframe>`;
                const frame = win.document.querySelector('iframe');
                if (!frame.contentDocument || frame.contentDocument.readyState !== 'complete')
                await new Promise(f => frame.addEventListener('load', f));
                // 3. After that, remove the iframe.
                frame.remove();
            }"""
        )
    popup = await page_info.value
    # 4. Connect to the popup and make sure it doesn't throw.
    assert await popup.evaluate("1 + 1") == 2



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import Dict

from playwright.async_api import Browser, BrowserType
from tests.server import Server
from tests.utils import must


async def test_should_construct_a_new_url_when_a_base_url_in_browser_new_context_is_passed(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(base_url=server.PREFIX)
    page = await context.new_page()
    assert (must(await page.goto("/empty.html"))).url == server.EMPTY_PAGE
    await context.close()


async def test_should_construct_a_new_url_when_a_base_url_in_browser_new_page_is_passed(
    browser: Browser, server: Server
) -> None:
    page = await browser.new_page(base_url=server.PREFIX)
    assert (must(await page.goto("/empty.html"))).url == server.EMPTY_PAGE
    await page.close()


async def test_should_construct_a_new_url_when_a_base_url_in_browser_new_persistent_context_is_passed(
    browser_type: BrowserType, tmp_path: Path, server: Server, launch_arguments: Dict
) -> None:
    context = await browser_type.launch_persistent_context(
        tmp_path, **launch_arguments, base_url=server.PREFIX
    )
    page = await context.new_page()
    assert (must(await page.goto("/empty.html"))).url == server.EMPTY_PAGE
    await context.close()


async def test_should_construct_correctly_when_a_baseurl_without_a_trailing_slash_is_passed(
    browser: Browser, server: Server
) -> None:
    page = await browser.new_page(base_url=server.PREFIX + "/url-construction")
    assert (must(await page.goto("mypage.html"))).url == server.PREFIX + "/mypage.html"
    assert (
        must(await page.goto("./mypage.html"))
    ).url == server.PREFIX + "/mypage.html"
    assert (must(await page.goto("/mypage.html"))).url == server.PREFIX + "/mypage.html"
    await page.close()


async def test_should_construct_correctly_when_a_baseurl_with_a_trailing_slash_is_passed(
    browser: Browser, server: Server
) -> None:
    page = await browser.new_page(base_url=server.PREFIX + "/url-construction/")
    assert (
        must(await page.goto("mypage.html"))
    ).url == server.PREFIX + "/url-construction/mypage.html"
    assert (
        must(await page.goto("./mypage.html"))
    ).url == server.PREFIX + "/url-construction/mypage.html"
    assert (must(await page.goto("/mypage.html"))).url == server.PREFIX + "/mypage.html"
    assert (must(await page.goto("."))).url == server.PREFIX + "/url-construction/"
    assert (must(await page.goto("/"))).url == server.PREFIX + "/"
    await page.close()


async def test_should_not_construct_a_new_url_when_valid_urls_are_passed(
    browser: Browser, server: Server
) -> None:
    page = await browser.new_page(base_url="http://microsoft.com")
    response = await page.goto(server.EMPTY_PAGE)
    assert response
    assert response.url == server.EMPTY_PAGE

    await page.goto("data:text/html,Hello world")
    assert page.url == "data:text/html,Hello world"

    await page.goto("about:blank")
    assert page.url == "about:blank"

    await page.close()


async def test_should_be_able_to_match_a_url_relative_to_its_given_url_with_urlmatcher(
    browser: Browser, server: Server
) -> None:
    page = await browser.new_page(base_url=server.PREFIX + "/foobar/")

    await page.goto("/kek/index.html")
    await page.wait_for_url("/kek/index.html")
    assert page.url == server.PREFIX + "/kek/index.html"

    await page.route(
        "./kek/index.html", lambda route: route.fulfill(body="base-url-matched-route")
    )

    async with page.expect_request("./kek/index.html") as request_info:
        async with page.expect_response("./kek/index.html") as response_info:
            await page.goto("./kek/index.html")
    request = await request_info.value
    response = await response_info.value
    assert request.url == server.PREFIX + "/foobar/kek/index.html"
    assert response.url == server.PREFIX + "/foobar/kek/index.html"
    assert await response.body() == b"base-url-matched-route"

    await page.close()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

import pytest

from playwright.sync_api import Page


@pytest.mark.only_browser("chromium")
def test_should_be_able_to_save_pdf_file(page: Page, tmp_path: Path) -> None:
    output_file = tmp_path / "foo.png"
    page.pdf(path=str(output_file))
    assert os.path.getsize(output_file) > 0


@pytest.mark.only_browser("chromium")
def test_should_be_able_capture_pdf_without_path(page: Page) -> None:
    buffer = page.pdf()
    assert buffer



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os
import platform
import shutil
import subprocess
import sys
import zipfile
from typing import Dict

driver_version = "1.55.0-beta-1756314050000"

base_wheel_bundles = [
    {
        "wheel": "macosx_10_13_x86_64.whl",
        "machine": "x86_64",
        "platform": "darwin",
        "zip_name": "mac",
    },
    {
        "wheel": "macosx_11_0_universal2.whl",
        "machine": "x86_64",
        "platform": "darwin",
        "zip_name": "mac",
    },
    {
        "wheel": "macosx_11_0_arm64.whl",
        "machine": "arm64",
        "platform": "darwin",
        "zip_name": "mac-arm64",
    },
    {
        "wheel": "manylinux1_x86_64.whl",
        "machine": "x86_64",
        "platform": "linux",
        "zip_name": "linux",
    },
    {
        "wheel": "manylinux_2_17_aarch64.manylinux2014_aarch64.whl",
        "machine": "aarch64",
        "platform": "linux",
        "zip_name": "linux-arm64",
    },
    {
        "wheel": "win32.whl",
        "machine": "i386",
        "platform": "win32",
        "zip_name": "win32_x64",
    },
    {
        "wheel": "win_amd64.whl",
        "machine": "amd64",
        "platform": "win32",
        "zip_name": "win32_x64",
    },
    {
        "wheel": "win_arm64.whl",
        "machine": "arm64",
        "platform": "win32",
        "zip_name": "win32_arm64",
    },
]

if len(sys.argv) == 2 and sys.argv[1] == "--list-wheels":
    for bundle in base_wheel_bundles:
        print(bundle["wheel"])
    exit(0)

from setuptools import setup  # noqa: E402

try:
    from auditwheel.wheeltools import InWheel
except ImportError:
    InWheel = None
from wheel.bdist_wheel import bdist_wheel as BDistWheelCommand  # noqa: E402


def extractall(zip: zipfile.ZipFile, path: str) -> None:
    for name in zip.namelist():
        member = zip.getinfo(name)
        extracted_path = zip.extract(member, path)
        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(extracted_path, attr)


def download_driver(zip_name: str) -> None:
    zip_file = f"playwright-{driver_version}-{zip_name}.zip"
    destination_path = "driver/" + zip_file
    if os.path.exists(destination_path):
        return
    url = "https://playwright.azureedge.net/builds/driver/"
    if (
        "-alpha" in driver_version
        or "-beta" in driver_version
        or "-next" in driver_version
    ):
        url = url + "next/"
    url = url + zip_file
    temp_destination_path = destination_path + ".tmp"
    print(f"Fetching {url}")
    # Don't replace this with urllib - Python won't have certificates to do SSL on all platforms.
    subprocess.check_call(["curl", url, "-o", temp_destination_path])
    os.rename(temp_destination_path, destination_path)


class PlaywrightBDistWheelCommand(BDistWheelCommand):
    def run(self) -> None:
        super().run()
        os.makedirs("driver", exist_ok=True)
        os.makedirs("playwright/driver", exist_ok=True)
        self._download_and_extract_local_driver()

        wheel = None
        if os.getenv("PLAYWRIGHT_TARGET_WHEEL", None):
            wheel = list(
                filter(
                    lambda wheel: wheel["wheel"]
                    == os.getenv("PLAYWRIGHT_TARGET_WHEEL"),
                    base_wheel_bundles,
                )
            )[0]
        else:
            wheel = list(
                filter(
                    lambda wheel: wheel["platform"] == sys.platform
                    and wheel["machine"] == platform.machine().lower(),
                    base_wheel_bundles,
                )
            )[0]
        assert wheel
        self._build_wheel(wheel)

    def _build_wheel(
        self,
        wheel_bundle: Dict[str, str],
    ) -> None:
        assert self.dist_dir
        base_wheel_location: str = glob.glob(os.path.join(self.dist_dir, "*.whl"))[0]
        without_platform = base_wheel_location[:-7]
        download_driver(wheel_bundle["zip_name"])
        zip_file = f"driver/playwright-{driver_version}-{wheel_bundle['zip_name']}.zip"
        with zipfile.ZipFile(zip_file, "r") as zip:
            extractall(zip, f"driver/{wheel_bundle['zip_name']}")
        wheel_location = without_platform + wheel_bundle["wheel"]
        shutil.copy(base_wheel_location, wheel_location)
        with zipfile.ZipFile(
            wheel_location, mode="a", compression=zipfile.ZIP_DEFLATED
        ) as zip:
            driver_root = os.path.abspath(f"driver/{wheel_bundle['zip_name']}")
            for dir_path, _, files in os.walk(driver_root):
                for file in files:
                    from_path = os.path.join(dir_path, file)
                    to_path = os.path.relpath(from_path, driver_root)
                    zip.write(from_path, f"playwright/driver/{to_path}")
            zip.writestr(
                "playwright/driver/README.md",
                f"{wheel_bundle['wheel']} driver package",
            )
        os.remove(base_wheel_location)
        for whlfile in glob.glob(os.path.join(self.dist_dir, "*.whl")):
            os.makedirs("wheelhouse", exist_ok=True)
            if InWheel:
                wheelhouse_whl = os.path.join("wheelhouse", os.path.basename(whlfile))
                shutil.move(whlfile, wheelhouse_whl)
                with InWheel(in_wheel=wheelhouse_whl, out_wheel=whlfile):
                    print(f"Updating RECORD file of {whlfile}")
        print("Copying new wheels")
        shutil.rmtree("wheelhouse")

    def _download_and_extract_local_driver(
        self,
    ) -> None:
        zip_names_for_current_system = set(
            map(
                lambda wheel: wheel["zip_name"],
                filter(
                    lambda wheel: wheel["machine"] == platform.machine().lower()
                    and wheel["platform"] == sys.platform,
                    base_wheel_bundles,
                ),
            )
        )
        assert len(zip_names_for_current_system) == 1
        zip_name = zip_names_for_current_system.pop()
        download_driver(zip_name)
        zip_file = f"driver/playwright-{driver_version}-{zip_name}.zip"
        with zipfile.ZipFile(zip_file, "r") as zip:
            extractall(zip, "playwright/driver")


setup(
    cmdclass={"bdist_wheel": PlaywrightBDistWheelCommand},
)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import math
from datetime import datetime, timezone
from typing import Any, Dict

from playwright.async_api import Page


async def test_jshandle_evaluate_work(page: Page) -> None:
    window_handle = await page.evaluate_handle("window")
    assert window_handle
    assert (
        repr(window_handle) == f"<JSHandle preview={window_handle._impl_obj._preview}>"
    )


async def test_jshandle_evaluate_accept_object_handle_as_argument(page: Page) -> None:
    navigator_handle = await page.evaluate_handle("navigator")
    text = await page.evaluate("e => e.userAgent", navigator_handle)
    assert "Mozilla" in text


async def test_jshandle_evaluate_accept_handle_to_primitive_types(page: Page) -> None:
    handle = await page.evaluate_handle("5")
    is_five = await page.evaluate("e => Object.is(e, 5)", handle)
    assert is_five


async def test_jshandle_evaluate_accept_nested_handle(page: Page) -> None:
    foo = await page.evaluate_handle('({ x: 1, y: "foo" })')
    result = await page.evaluate("({ foo }) => foo", {"foo": foo})
    assert result == {"x": 1, "y": "foo"}


async def test_jshandle_evaluate_accept_nested_window_handle(page: Page) -> None:
    foo = await page.evaluate_handle("window")
    result = await page.evaluate("({ foo }) => foo === window", {"foo": foo})
    assert result


async def test_jshandle_evaluate_accept_multiple_nested_handles(page: Page) -> None:
    foo = await page.evaluate_handle('({ x: 1, y: "foo" })')
    bar = await page.evaluate_handle("5")
    baz = await page.evaluate_handle('["baz"]')
    result = await page.evaluate(
        "x => JSON.stringify(x)",
        {"a1": {"foo": foo}, "a2": {"bar": bar, "arr": [{"baz": baz}]}},
    )
    assert json.loads(result) == {
        "a1": {"foo": {"x": 1, "y": "foo"}},
        "a2": {"bar": 5, "arr": [{"baz": ["baz"]}]},
    }


async def test_jshandle_evaluate_should_work_for_circular_objects(page: Page) -> None:
    a: Dict[str, Any] = {"x": 1}
    a["y"] = a
    result = await page.evaluate("a => { a.y.x += 1; return a; }", a)
    assert result["x"] == 2
    assert result["y"]["x"] == 2
    assert result == result["y"]


async def test_jshandle_evaluate_accept_same_nested_object_multiple_times(
    page: Page,
) -> None:
    foo = {"x": 1}
    assert await page.evaluate(
        "x => x", {"foo": foo, "bar": [foo], "baz": {"foo": foo}}
    ) == {"foo": {"x": 1}, "bar": [{"x": 1}], "baz": {"foo": {"x": 1}}}


async def test_jshandle_evaluate_accept_object_handle_to_unserializable_value(
    page: Page,
) -> None:
    handle = await page.evaluate_handle("() => Infinity")
    assert await page.evaluate("e => Object.is(e, Infinity)", handle)


async def test_jshandle_evaluate_pass_configurable_args(page: Page) -> None:
    result = await page.evaluate(
        """arg => {
            if (arg.foo !== 42)
            throw new Error('Not a 42');
            arg.foo = 17;
            if (arg.foo !== 17)
            throw new Error('Not 17');
            delete arg.foo;
            if (arg.foo === 17)
            throw new Error('Still 17');
            return arg;
        }""",
        {"foo": 42},
    )
    assert result == {}


async def test_jshandle_properties_get_property(page: Page) -> None:
    handle1 = await page.evaluate_handle(
        """() => ({
            one: 1,
            two: 2,
            three: 3
        })"""
    )
    handle2 = await handle1.get_property("two")
    assert await handle2.json_value() == 2


async def test_jshandle_properties_work_with_undefined_null_and_empty(
    page: Page,
) -> None:
    handle = await page.evaluate_handle(
        """() => ({
            undefined: undefined,
            null: null,
        })"""
    )
    undefined_handle = await handle.get_property("undefined")
    assert await undefined_handle.json_value() is None
    null_handle = await handle.get_property("null")
    assert await null_handle.json_value() is None
    empty_handle = await handle.get_property("empty")
    assert await empty_handle.json_value() is None


async def test_jshandle_properties_work_with_unserializable_values(page: Page) -> None:
    handle = await page.evaluate_handle(
        """() => ({
            infinity: Infinity,
            negInfinity: -Infinity,
            nan: NaN,
            negZero: -0
        })"""
    )
    infinity_handle = await handle.get_property("infinity")
    assert await infinity_handle.json_value() == float("inf")
    neg_infinity_handle = await handle.get_property("negInfinity")
    assert await neg_infinity_handle.json_value() == float("-inf")
    nan_handle = await handle.get_property("nan")
    assert math.isnan(await nan_handle.json_value()) is True
    neg_zero_handle = await handle.get_property("negZero")
    assert await neg_zero_handle.json_value() == float("-0")


async def test_jshandle_properties_get_properties(page: Page) -> None:
    handle = await page.evaluate_handle('() => ({ foo: "bar" })')
    properties = await handle.get_properties()
    assert "foo" in properties
    foo = properties["foo"]
    assert await foo.json_value() == "bar"


async def test_jshandle_properties_return_empty_map_for_non_objects(page: Page) -> None:
    handle = await page.evaluate_handle("123")
    properties = await handle.get_properties()
    assert properties == {}


async def test_jshandle_json_value_work(page: Page) -> None:
    handle = await page.evaluate_handle('() => ({foo: "bar"})')
    json = await handle.json_value()
    assert json == {"foo": "bar"}


async def test_jshandle_json_value_work_with_dates(page: Page) -> None:
    handle = await page.evaluate_handle('() => new Date("2020-05-27T01:31:38.506Z")')
    json = await handle.json_value()
    assert json == datetime.fromisoformat("2020-05-27T01:31:38.506").replace(
        tzinfo=timezone.utc
    )


async def test_jshandle_json_value_should_work_for_circular_object(page: Page) -> None:
    handle = await page.evaluate_handle("const a = {}; a.b = a; a")
    a: Dict[str, Any] = {}
    a["b"] = a
    result = await handle.json_value()
    # Node test looks like the below, but assert isn't smart enough to handle this:
    # assert await handle.json_value() == a
    assert result["b"] == result


async def test_jshandle_as_element_work(page: Page) -> None:
    handle = await page.evaluate_handle("document.body")
    element = handle.as_element()
    assert element is not None


async def test_jshandle_as_element_return_none_for_non_elements(page: Page) -> None:
    handle = await page.evaluate_handle("2")
    element = handle.as_element()
    assert element is None


async def test_jshandle_to_string_work_for_primitives(page: Page) -> None:
    number_handle = await page.evaluate_handle("2")
    assert str(number_handle) == "2"
    string_handle = await page.evaluate_handle('"a"')
    assert str(string_handle) == "a"


async def test_jshandle_to_string_work_for_complicated_objects(
    page: Page, browser_name: str
) -> None:
    handle = await page.evaluate_handle("window")
    if browser_name != "firefox":
        assert str(handle) == "Window"
    else:
        assert str(handle) == "JSHandle@object"


async def test_jshandle_to_string_work_for_promises(page: Page) -> None:
    handle = await page.evaluate_handle("({b: Promise.resolve(123)})")
    b_handle = await handle.get_property("b")
    assert str(b_handle) == "Promise"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import traceback
from typing import Callable
from urllib.parse import urlparse

import pytest

from playwright._impl._path_utils import get_file_dirname
from playwright.sync_api import Error, Page, expect
from tests.server import Server

_dirname = get_file_dirname()
FILE_TO_UPLOAD = _dirname / ".." / "assets/file-to-upload.txt"


def test_locators_click_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/button.html")
    button = page.locator("button")
    button.click()
    assert page.evaluate("window['result']") == "Clicked"


def test_locators_click_should_work_with_node_removed(
    page: Page, server: Server
) -> None:
    page.goto(server.PREFIX + "/input/button.html")
    page.evaluate("delete window['Node']")
    button = page.locator("button")
    button.click()
    assert page.evaluate("window['result']") == "Clicked"


def test_locators_click_should_work_for_text_nodes(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/button.html")
    page.evaluate(
        """() => {
        window['double'] = false;
        const button = document.querySelector('button');
        button.addEventListener('dblclick', event => {
        window['double'] = true;
        });
    }"""
    )
    button = page.locator("button")
    button.dblclick()
    assert page.evaluate("double") is True
    assert page.evaluate("result") == "Clicked"


def test_locators_should_have_repr(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/button.html")
    button = page.locator("button")
    button.click()
    assert (
        str(button)
        == f"<Locator frame=<Frame name= url='{server.PREFIX}/input/button.html'> selector='button'>"
    )


def test_locators_get_attribute_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/dom.html")
    button = page.locator("#outer")
    assert button.get_attribute("name") == "value"
    assert button.get_attribute("foo") is None


def test_locators_input_value_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/dom.html")
    page.fill("#textarea", "input value")
    text_area = page.locator("#textarea")
    assert text_area.input_value() == "input value"


def test_locators_inner_html_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/dom.html")
    locator = page.locator("#outer")
    assert locator.inner_html() == '<div id="inner">Text,\nmore text</div>'


def test_locators_inner_text_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/dom.html")
    locator = page.locator("#inner")
    assert locator.inner_text() == "Text, more text"


def test_locators_text_content_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/dom.html")
    locator = page.locator("#inner")
    assert locator.text_content() == "Text,\nmore text"


def test_locators_is_hidden_and_is_visible_should_work(page: Page) -> None:
    page.set_content("<div>Hi</div><span></span>")

    div = page.locator("div")
    assert div.is_visible() is True
    assert div.is_hidden() is False

    span = page.locator("span")
    assert span.is_visible() is False
    assert span.is_hidden() is True


def test_locators_is_enabled_and_is_disabled_should_work(page: Page) -> None:
    page.set_content(
        """
        <button disabled>button1</button>
        <button>button2</button>
        <div>div</div>
    """
    )

    div = page.locator("div")
    assert div.is_enabled()
    assert div.is_disabled() is False

    button1 = page.locator(':text("button1")')
    assert button1.is_enabled() is False
    assert button1.is_disabled() is True

    button1 = page.locator(':text("button2")')
    assert button1.is_enabled()
    assert button1.is_disabled() is False


def test_locators_is_editable_should_work(page: Page) -> None:
    page.set_content(
        """
        <input id=input1 disabled><textarea></textarea><input id=input2>
    """
    )

    input1 = page.locator("#input1")
    assert input1.is_editable() is False

    input2 = page.locator("#input2")
    assert input2.is_editable() is True


def test_locators_is_checked_should_work(page: Page) -> None:
    page.set_content(
        """
        <input type='checkbox' checked><div>Not a checkbox</div>
    """
    )

    element = page.locator("input")
    assert element.is_checked() is True
    element.evaluate("e => e.checked = false")
    assert element.is_checked() is False


def test_locators_all_text_contents_should_work(page: Page) -> None:
    page.set_content(
        """
        <div>A</div><div>B</div><div>C</div>
    """
    )

    element = page.locator("div")
    assert element.all_text_contents() == ["A", "B", "C"]


def test_locators_all_inner_texts(page: Page) -> None:
    page.set_content(
        """
        <div>A</div><div>B</div><div>C</div>
    """
    )

    element = page.locator("div")
    assert element.all_inner_texts() == ["A", "B", "C"]


def test_locators_should_query_existing_element(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/playground.html")
    page.set_content(
        """<html><body><div class="second"><div class="inner">A</div></div></body></html>"""
    )
    html = page.locator("html")
    second = html.locator(".second")
    inner = second.locator(".inner")
    assert page.evaluate("e => e.textContent", inner.element_handle()) == "A"


def test_locators_evaluate_handle_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/dom.html")
    outer = page.locator("#outer")
    inner = outer.locator("#inner")
    check = inner.locator("#check")
    text = inner.evaluate_handle("e => e.firstChild")
    page.evaluate("1 + 1")
    assert (
        str(outer)
        == f"<Locator frame=<Frame name= url='{server.PREFIX}/dom.html'> selector='#outer'>"
    )
    assert (
        str(inner)
        == f"<Locator frame=<Frame name= url='{server.PREFIX}/dom.html'> selector='#outer >> #inner'>"
    )
    assert str(text) == "JSHandle@#text=Text,↵more text"
    assert (
        str(check)
        == f"<Locator frame=<Frame name= url='{server.PREFIX}/dom.html'> selector='#outer >> #inner >> #check'>"
    )


def test_locators_should_query_existing_elements(page: Page) -> None:
    page.set_content("""<html><body><div>A</div><br/><div>B</div></body></html>""")
    html = page.locator("html")
    elements = html.locator("div").element_handles()
    assert len(elements) == 2
    result = []
    for element in elements:
        result.append(page.evaluate("e => e.textContent", element))
    assert result == ["A", "B"]


def test_locators_return_empty_array_for_non_existing_elements(page: Page) -> None:
    page.set_content("""<html><body><div>A</div><br/><div>B</div></body></html>""")
    html = page.locator("html")
    elements = html.locator("abc").element_handles()
    assert len(elements) == 0
    assert elements == []


def test_locators_evaluate_all_should_work(page: Page) -> None:
    page.set_content(
        """<html><body><div class="tweet"><div class="like">100</div><div class="like">10</div></div></body></html>"""
    )
    tweet = page.locator(".tweet .like")
    content = tweet.evaluate_all("nodes => nodes.map(n => n.innerText)")
    assert content == ["100", "10"]


def test_locators_evaluate_all_should_work_with_missing_selector(page: Page) -> None:
    page.set_content("""<div class="a">not-a-child-div</div><div id="myId"></div""")
    tweet = page.locator("#myId .a")
    nodes_length = tweet.evaluate_all("nodes => nodes.length")
    assert nodes_length == 0


def test_locators_hover_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/scrollable.html")
    button = page.locator("#button-6")
    button.hover()
    assert page.evaluate("document.querySelector('button:hover').id") == "button-6"


def test_locators_fill_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/textarea.html")
    button = page.locator("input")
    button.fill("some value")
    assert page.evaluate("result") == "some value"


def test_locators_clear_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/textarea.html")
    button = page.locator("input")
    button.fill("some value")
    assert page.evaluate("result") == "some value"
    button.clear()
    assert page.evaluate("result") == ""


def test_locators_check_should_work(page: Page) -> None:
    page.set_content("<input id='checkbox' type='checkbox'></input>")
    button = page.locator("input")
    button.check()
    assert page.evaluate("checkbox.checked") is True


def test_locators_uncheck_should_work(page: Page) -> None:
    page.set_content("<input id='checkbox' type='checkbox' checked></input>")
    button = page.locator("input")
    button.uncheck()
    assert page.evaluate("checkbox.checked") is False


def test_locators_select_option_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/select.html")
    select = page.locator("select")
    select.select_option("blue")
    assert page.evaluate("result.onInput") == ["blue"]
    assert page.evaluate("result.onChange") == ["blue"]


def test_locators_focus_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/button.html")
    button = page.locator("button")
    assert button.evaluate("button => document.activeElement === button") is False
    button.focus()
    assert button.evaluate("button => document.activeElement === button") is True


def test_locators_dispatch_event_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/button.html")
    button = page.locator("button")
    button.dispatch_event("click")
    assert page.evaluate("result") == "Clicked"


def test_locators_should_upload_a_file(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/fileupload.html")
    input = page.locator("input[type=file]")

    file_path = os.path.relpath(FILE_TO_UPLOAD, os.getcwd())
    input.set_input_files(file_path)
    assert (
        page.evaluate("e => e.files[0].name", input.element_handle())
        == "file-to-upload.txt"
    )


def test_locators_should_press(page: Page) -> None:
    page.set_content("<input type='text' />")
    page.locator("input").press("h")
    assert page.eval_on_selector("input", "input => input.value") == "h"


def test_locators_should_scroll_into_view(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/offscreenbuttons.html")
    for i in range(11):
        button = page.locator(f"#btn{i}")
        before = button.evaluate(
            "button => button.getBoundingClientRect().right - window.innerWidth"
        )
        assert before == 10 * i
        button.scroll_into_view_if_needed()
        after = button.evaluate(
            "button => button.getBoundingClientRect().right - window.innerWidth"
        )
        assert after <= 0
        page.evaluate("window.scrollTo(0, 0)")


def test_locators_should_select_textarea(
    page: Page, server: Server, browser_name: str
) -> None:
    page.goto(server.PREFIX + "/input/textarea.html")
    textarea = page.locator("textarea")
    textarea.evaluate("textarea => textarea.value = 'some value'")
    textarea.select_text()
    textarea.select_text(timeout=25_000)
    if browser_name == "firefox" or browser_name == "webkit":
        assert textarea.evaluate("el => el.selectionStart") == 0
        assert textarea.evaluate("el => el.selectionEnd") == 10
    else:
        assert page.evaluate("window.getSelection().toString()") == "some value"


def test_locators_should_type(page: Page) -> None:
    page.set_content("<input type='text' />")
    page.locator("input").type("hello")
    assert page.eval_on_selector("input", "input => input.value") == "hello"


def test_locators_should_screenshot(
    page: Page, server: Server, assert_to_be_golden: Callable[[bytes, str], None]
) -> None:
    page.set_viewport_size(
        {
            "width": 500,
            "height": 500,
        }
    )
    page.goto(server.PREFIX + "/grid.html")
    page.evaluate("window.scrollBy(50, 100)")
    element = page.locator(".box:nth-of-type(3)")
    assert_to_be_golden(element.screenshot(), "screenshot-element-bounding-box.png")
    assert_to_be_golden(
        element.screenshot(timeout=1_000), "screenshot-element-bounding-box.png"
    )


def test_locators_should_return_bounding_box(page: Page, server: Server) -> None:
    page.set_viewport_size(
        {
            "width": 500,
            "height": 500,
        }
    )
    page.goto(server.PREFIX + "/grid.html")
    element = page.locator(".box:nth-of-type(13)")
    box = element.bounding_box()
    assert box == {
        "x": 100,
        "y": 50,
        "width": 50,
        "height": 50,
    }


def test_locators_should_respect_first_and_last(page: Page) -> None:
    page.set_content(
        """
        <section>
            <div><p>A</p></div>
            <div><p>A</p><p>A</p></div>
            <div><p>A</p><p>A</p><p>A</p></div>
        </section>"""
    )
    assert page.locator("div >> p").count() == 6
    assert page.locator("div").locator("p").count() == 6
    assert page.locator("div").first.locator("p").count() == 1
    assert page.locator("div").last.locator("p").count() == 3


def test_locators_should_respect_nth(page: Page) -> None:
    page.set_content(
        """
    <section>
        <div><p>A</p></div>
        <div><p>A</p><p>A</p></div>
        <div><p>A</p><p>A</p><p>A</p></div>
    </section>"""
    )
    assert page.locator("div >> p").nth(0).count() == 1
    assert page.locator("div").nth(1).locator("p").count() == 2
    assert page.locator("div").nth(2).locator("p").count() == 3


def test_locators_should_throw_on_capture_without_nth(page: Page) -> None:
    page.set_content(
        """
        <section><div><p>A</p></div></section>
    """
    )
    with pytest.raises(Error, match="Can't query n-th element"):
        page.locator("*css=div >> p").nth(1).click()


def test_locators_should_throw_due_to_strictness(page: Page) -> None:
    page.set_content(
        """
        <div>A</div><div>B</div>
    """
    )
    with pytest.raises(Error, match="strict mode violation"):
        page.locator("div").is_visible()


def test_locators_should_throw_due_to_strictness_2(page: Page) -> None:
    page.set_content(
        """
        <select><option>One</option><option>Two</option></select>
    """
    )
    with pytest.raises(Error, match="strict mode violation"):
        page.locator("option").evaluate("e => {}")


def test_locators_set_checked(page: Page) -> None:
    page.set_content("`<input id='checkbox' type='checkbox'></input>`")
    locator = page.locator("input")
    locator.set_checked(True)
    assert page.evaluate("checkbox.checked")
    locator.set_checked(False)
    assert page.evaluate("checkbox.checked") is False


def test_should_combine_visible_with_other_selectors(page: Page) -> None:
    page.set_content(
        """<div>
        <div class="item" style="display: none">Hidden data0</div>
        <div class="item">visible data1</div>
        <div class="item" style="display: none">Hidden data1</div>
        <div class="item">visible data2</div>
        <div class="item" style="display: none">Hidden data1</div>
        <div class="item">visible data3</div>
        </div>
    """
    )
    locator = page.locator(".item >> visible=true").nth(1)
    expect(locator).to_have_text("visible data2")
    expect(page.locator(".item >> visible=true >> text=data3")).to_have_text(
        "visible data3"
    )


def test_should_support_filter_visible(page: Page) -> None:
    page.set_content(
        """<div>
    <div class="item" style="display: none">Hidden data0</div>
    <div class="item">visible data1</div>
    <div class="item" style="display: none">Hidden data1</div>
    <div class="item">visible data2</div>
    <div class="item" style="display: none">Hidden data2</div>
    <div class="item">visible data3</div>
    </div>
    """
    )
    locator = page.locator(".item").filter(visible=True).nth(1)
    expect(locator).to_have_text("visible data2")
    expect(
        page.locator(".item").filter(visible=True).get_by_text("data3")
    ).to_have_text("visible data3")
    expect(
        page.locator(".item").filter(visible=False).get_by_text("data1")
    ).to_have_text("Hidden data1")


def test_locator_count_should_work_with_deleted_map_in_main_world(page: Page) -> None:
    page.evaluate("Map = 1")
    page.locator("#searchResultTableDiv .x-grid3-row").count()
    expect(page.locator("#searchResultTableDiv .x-grid3-row")).to_have_count(0)


def test_locator_locator_and_framelocator_locator_should_accept_locator(
    page: Page,
) -> None:
    page.set_content(
        """
        <div><input value=outer></div>
        <iframe srcdoc="<div><input value=inner></div>"></iframe>
    """
    )

    input_locator = page.locator("input")
    assert input_locator.input_value() == "outer"
    assert page.locator("div").locator(input_locator).input_value() == "outer"
    assert page.frame_locator("iframe").locator(input_locator).input_value() == "inner"
    assert (
        page.frame_locator("iframe").locator("div").locator(input_locator).input_value()
        == "inner"
    )

    div_locator = page.locator("div")
    assert div_locator.locator("input").input_value() == "outer"
    assert (
        page.frame_locator("iframe").locator(div_locator).locator("input").input_value()
        == "inner"
    )


def route_iframe(page: Page) -> None:
    page.route(
        "**/empty.html",
        lambda route: route.fulfill(
            body='<iframe src="iframe.html" name="frame1"></iframe>',
            content_type="text/html",
        ),
    )
    page.route(
        "**/iframe.html",
        lambda route: route.fulfill(
            body="""<html>
          <div>
            <button>Hello iframe</button>
            <iframe src="iframe-2.html"></iframe>
          </div>
          <span>1</span>
          <span>2</span>
        </html>""",
            content_type="text/html",
        ),
    )
    page.route(
        "**/iframe-2.html",
        lambda route: route.fulfill(
            body="<html><button>Hello nested iframe</button></html>",
            content_type="text/html",
        ),
    )


def test_locators_frame_should_work_with_iframe(page: Page, server: Server) -> None:
    route_iframe(page)
    page.goto(server.EMPTY_PAGE)
    button = page.frame_locator("iframe").locator("button")
    button.wait_for()
    assert button.inner_text() == "Hello iframe"
    button.click()
    assert (
        repr(page.frame_locator("iframe"))
        == f"<FrameLocator frame=<Frame name= url='{server.PREFIX}/empty.html'> selector='iframe'>"
    )


def test_locators_frame_should_work_for_nested_iframe(
    page: Page, server: Server
) -> None:
    route_iframe(page)
    page.goto(server.EMPTY_PAGE)
    button = page.frame_locator("iframe").frame_locator("iframe").locator("button")
    button.wait_for()
    assert button.inner_text() == "Hello nested iframe"
    button.click()


def test_locators_frame_should_work_with_locator_frame_locator(
    page: Page, server: Server
) -> None:
    route_iframe(page)
    page.goto(server.EMPTY_PAGE)
    button = page.locator("body").frame_locator("iframe").locator("button")
    button.wait_for()
    assert button.inner_text() == "Hello iframe"
    button.click()


def test_locator_content_frame_should_work(page: Page, server: Server) -> None:
    route_iframe(page)
    page.goto(server.EMPTY_PAGE)
    locator = page.locator("iframe")
    frame_locator = locator.content_frame
    button = frame_locator.locator("button")
    assert button.inner_text() == "Hello iframe"
    expect(button).to_have_text("Hello iframe")
    button.click()


def test_frame_locator_owner_should_work(page: Page, server: Server) -> None:
    route_iframe(page)
    page.goto(server.EMPTY_PAGE)
    frame_locator = page.frame_locator("iframe")
    locator = frame_locator.owner
    expect(locator).to_be_visible()
    assert locator.get_attribute("name") == "frame1"


def route_ambiguous(page: Page) -> None:
    page.route(
        "**/empty.html",
        lambda route: route.fulfill(
            body="""
        <iframe src="iframe-1.html"></iframe>
        <iframe src="iframe-2.html"></iframe>
        <iframe src="iframe-3.html"></iframe>
    """,
            content_type="text/html",
        ),
    )
    page.route(
        "**/iframe-*",
        lambda route: route.fulfill(
            body=f"<html><button>Hello from {urlparse(route.request.url).path[1:]}</button></html>",
            content_type="text/html",
        ),
    )


def test_locator_frame_locator_should_throw_on_ambiguity(
    page: Page, server: Server
) -> None:
    route_ambiguous(page)
    page.goto(server.EMPTY_PAGE)
    button = page.locator("body").frame_locator("iframe").locator("button")
    with pytest.raises(
        Error,
        match=r'.*strict mode violation: locator\("body"\)\.locator\("iframe"\) resolved to 3 elements.*',
    ):
        button.wait_for()


def test_locator_frame_locator_should_not_throw_on_first_last_nth(
    page: Page, server: Server
) -> None:
    route_ambiguous(page)
    page.goto(server.EMPTY_PAGE)
    button1 = page.locator("body").frame_locator("iframe").first.locator("button")
    assert button1.text_content() == "Hello from iframe-1.html"
    button2 = page.locator("body").frame_locator("iframe").nth(1).locator("button")
    assert button2.text_content() == "Hello from iframe-2.html"
    button3 = page.locator("body").frame_locator("iframe").last.locator("button")
    assert button3.text_content() == "Hello from iframe-3.html"


def test_drag_to(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/drag-n-drop.html")
    page.locator("#source").drag_to(page.locator("#target"))
    assert (
        page.eval_on_selector(
            "#target", "target => target.contains(document.querySelector('#source'))"
        )
        is True
    )


def test_locator_query_should_filter_by_text(page: Page, server: Server) -> None:
    page.set_content("<div>Foobar</div><div>Bar</div>")
    expect(page.locator("div", has_text="Foo")).to_have_text("Foobar")


def test_locator_query_should_filter_by_text_2(page: Page, server: Server) -> None:
    page.set_content("<div>foo <span>hello world</span> bar</div>")
    expect(page.locator("div", has_text="hello world")).to_have_text(
        "foo hello world bar"
    )


def test_locator_query_should_filter_by_regex(page: Page, server: Server) -> None:
    page.set_content("<div>Foobar</div><div>Bar</div>")
    expect(page.locator("div", has_text=re.compile(r"Foo.*"))).to_have_text("Foobar")


def test_locator_query_should_filter_by_text_with_quotes(
    page: Page, server: Server
) -> None:
    page.set_content('<div>Hello "world"</div><div>Hello world</div>')
    expect(page.locator("div", has_text='Hello "world"')).to_have_text('Hello "world"')


def test_locator_query_should_filter_by_regex_with_quotes(
    page: Page, server: Server
) -> None:
    page.set_content('<div>Hello "world"</div><div>Hello world</div>')
    expect(page.locator("div", has_text=re.compile('Hello "world"'))).to_have_text(
        'Hello "world"'
    )


def test_locator_query_should_filter_by_regex_and_regexp_flags(
    page: Page, server: Server
) -> None:
    page.set_content('<div>Hello "world"</div><div>Hello world</div>')
    expect(
        page.locator("div", has_text=re.compile('hElLo "world', re.IGNORECASE))
    ).to_have_text('Hello "world"')


def test_locator_should_return_page(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/frames/two-frames.html")
    outer = page.locator("#outer")
    assert outer.page == page

    inner = outer.locator("#inner")
    assert inner.page == page

    in_frame = page.frames[1].locator("div")
    assert in_frame.page == page


def test_locator_should_support_has_locator(page: Page, server: Server) -> None:
    page.set_content("<div><span>hello</span></div><div><span>world</span></div>")
    expect(page.locator("div", has=page.locator("text=world"))).to_have_count(1)
    assert (
        page.locator("div", has=page.locator("text=world")).evaluate("e => e.outerHTML")
        == "<div><span>world</span></div>"
    )
    expect(page.locator("div", has=page.locator('text="hello"'))).to_have_count(1)
    assert (
        page.locator("div", has=page.locator('text="hello"')).evaluate(
            "e => e.outerHTML"
        )
        == "<div><span>hello</span></div>"
    )
    expect(page.locator("div", has=page.locator("xpath=./span"))).to_have_count(2)
    expect(page.locator("div", has=page.locator("span"))).to_have_count(2)
    expect(page.locator("div", has=page.locator("span", has_text="wor"))).to_have_count(
        1
    )
    assert (
        page.locator("div", has=page.locator("span", has_text="wor")).evaluate(
            "e => e.outerHTML"
        )
        == "<div><span>world</span></div>"
    )
    expect(
        page.locator(
            "div",
            has=page.locator("span"),
            has_text="wor",
        )
    ).to_have_count(1)


def test_locator_should_enforce_same_frame_for_has_locator(
    page: Page, server: Server
) -> None:
    page.goto(server.PREFIX + "/frames/two-frames.html")
    child = page.frames[1]
    with pytest.raises(Error) as exc_info:
        page.locator("div", has=child.locator("span"))
    assert (
        'Inner "has" locator must belong to the same frame.' in exc_info.value.message
    )


def test_locator_should_support_locator_or(page: Page, server: Server) -> None:
    page.set_content("<div>hello</div><span>world</span>")
    expect(page.locator("div").or_(page.locator("span"))).to_have_count(2)
    expect(page.locator("div").or_(page.locator("span"))).to_have_text(
        ["hello", "world"]
    )
    expect(
        page.locator("span").or_(page.locator("article")).or_(page.locator("div"))
    ).to_have_text(["hello", "world"])
    expect(page.locator("article").or_(page.locator("someting"))).to_have_count(0)
    expect(page.locator("article").or_(page.locator("div"))).to_have_text("hello")
    expect(page.locator("article").or_(page.locator("span"))).to_have_text("world")
    expect(page.locator("div").or_(page.locator("article"))).to_have_text("hello")
    expect(page.locator("span").or_(page.locator("article"))).to_have_text("world")


def test_locator_highlight_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/grid.html")
    page.locator(".box").nth(3).highlight()
    assert page.locator("x-pw-glass").is_visible()


def test_should_support_locator_that(page: Page) -> None:
    page.set_content(
        "<section><div><span>hello</span></div><div><span>world</span></div></section>"
    )

    expect(page.locator("div").filter(has_text="hello")).to_have_count(1)
    expect(
        page.locator("div", has_text="hello").filter(has_text="hello")
    ).to_have_count(1)
    expect(
        page.locator("div", has_text="hello").filter(has_text="world")
    ).to_have_count(0)
    expect(
        page.locator("section", has_text="hello").filter(has_text="world")
    ).to_have_count(1)
    expect(page.locator("div").filter(has_text="hello").locator("span")).to_have_count(
        1
    )
    expect(
        page.locator("div").filter(has=page.locator("span", has_text="world"))
    ).to_have_count(1)
    expect(page.locator("div").filter(has=page.locator("span"))).to_have_count(2)
    expect(
        page.locator("div").filter(
            has=page.locator("span"),
            has_text="world",
        )
    ).to_have_count(1)


def test_should_filter_by_case_insensitive_regex_in_a_child(page: Page) -> None:
    page.set_content('<div class="test"><h5>Title Text</h5></div>')
    expect(
        page.locator("div", has_text=re.compile(r"^title text$", re.I))
    ).to_have_text("Title Text")


def test_should_filter_by_case_insensitive_regex_in_multiple_children(
    page: Page,
) -> None:
    page.set_content('<div class="test"><h5>Title</h5> <h2><i>Text</i></h2></div>')
    expect(
        page.locator("div", has_text=re.compile(r"^title text$", re.I))
    ).to_have_class("test")


def test_should_filter_by_regex_with_special_symbols(page: Page) -> None:
    page.set_content(
        '<div class="test"><h5>First/"and"</h5><h2><i>Second\\</i></h2></div>'
    )
    expect(
        page.locator("div", has_text=re.compile(r'^first\/".*"second\\$', re.S | re.I))
    ).to_have_class("test")


def test_should_support_locator_filter(page: Page) -> None:
    page.set_content(
        "<section><div><span>hello</span></div><div><span>world</span></div></section>"
    )

    expect(page.locator("div").filter(has_text="hello")).to_have_count(1)
    expect(
        page.locator("div", has_text="hello").filter(has_text="hello")
    ).to_have_count(1)
    expect(
        page.locator("div", has_text="hello").filter(has_text="world")
    ).to_have_count(0)
    expect(
        page.locator("section", has_text="hello").filter(has_text="world")
    ).to_have_count(1)
    expect(page.locator("div").filter(has_text="hello").locator("span")).to_have_count(
        1
    )
    expect(
        page.locator("div").filter(has=page.locator("span", has_text="world"))
    ).to_have_count(1)
    expect(page.locator("div").filter(has=page.locator("span"))).to_have_count(2)
    expect(
        page.locator("div").filter(
            has=page.locator("span"),
            has_text="world",
        )
    ).to_have_count(1)
    expect(
        page.locator("div").filter(has_not=page.locator("span", has_text="world"))
    ).to_have_count(1)
    expect(page.locator("div").filter(has_not=page.locator("section"))).to_have_count(2)
    expect(page.locator("div").filter(has_not=page.locator("span"))).to_have_count(0)

    expect(page.locator("div").filter(has_not_text="hello")).to_have_count(1)
    expect(page.locator("div").filter(has_not_text="foo")).to_have_count(2)


def test_locators_should_support_locator_and(page: Page) -> None:
    page.set_content(
        """
        <div data-testid=foo>hello</div><div data-testid=bar>world</div>
        <span data-testid=foo>hello2</span><span data-testid=bar>world2</span>
    """
    )
    expect(page.locator("div").and_(page.locator("div"))).to_have_count(2)
    expect(page.locator("div").and_(page.get_by_test_id("foo"))).to_have_text(["hello"])
    expect(page.locator("div").and_(page.get_by_test_id("bar"))).to_have_text(["world"])
    expect(page.get_by_test_id("foo").and_(page.locator("div"))).to_have_text(["hello"])
    expect(page.get_by_test_id("bar").and_(page.locator("span"))).to_have_text(
        ["world2"]
    )
    expect(
        page.locator("span").and_(page.get_by_test_id(re.compile("bar|foo")))
    ).to_have_count(2)


def test_locators_has_does_not_encode_unicode(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    locators = [
        page.locator("button", has_text="Драматург"),
        page.locator("button", has_text=re.compile("Драматург")),
        page.locator("button", has=page.locator("text=Драматург")),
    ]
    for locator in locators:
        with pytest.raises(Error) as exc_info:
            locator.click(timeout=1_000)
        assert "Драматург" in exc_info.value.message


def test_locators_should_focus_and_blur_a_button(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/button.html")
    button = page.locator("button")
    assert not button.evaluate("button => document.activeElement === button")

    focused = False
    blurred = False

    def focus_event() -> None:
        nonlocal focused
        focused = True

    def blur_event() -> None:
        nonlocal blurred
        blurred = True

    page.expose_function("focusEvent", focus_event)
    page.expose_function("blurEvent", blur_event)
    button.evaluate(
        """button => {
        button.addEventListener('focus', window['focusEvent']);
        button.addEventListener('blur', window['blurEvent']);
    }"""
    )

    button.focus()
    assert focused
    assert not blurred
    assert button.evaluate("button => document.activeElement === button")

    button.blur()
    assert focused
    assert blurred
    assert not button.evaluate("button => document.activeElement === button")


def test_locator_all_should_work(page: Page) -> None:
    page.set_content("<div><p>A</p><p>B</p><p>C</p></div>")
    texts = []
    for p in page.locator("p").all():
        texts.append(p.text_content())
    assert texts == ["A", "B", "C"]


def test_locator_click_timeout_error_should_contain_call_log(page: Page) -> None:
    with pytest.raises(Error) as exc_info:
        page.get_by_role("button", name="Hello Python").click(timeout=42)
    formatted_exception = "".join(
        traceback.format_exception(type(exc_info.value), value=exc_info.value, tb=None)
    )
    assert "Locator.click: Timeout 42ms exceeded." in formatted_exception
    assert (
        'waiting for get_by_role("button", name="Hello Python")' in formatted_exception
    )
    assert (
        "During handling of the above exception, another exception occurred"
        not in formatted_exception
    )


def test_locator_should_ignore_deprecated_is_hidden_and_visible_timeout(
    page: Page,
) -> None:
    page.set_content("<div>foo</div>")
    div = page.locator("div")
    assert div.is_hidden(timeout=10) is False
    assert div.is_visible(timeout=10) is True



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from pathlib import Path

from playwright.sync_api import Playwright, sync_playwright


def main(playwright: Playwright, browser_name: str) -> None:
    browser = playwright[browser_name].launch()
    page = browser.new_page()
    page.goto("data:text/html,Foobar")
    here = Path(__file__).parent.resolve()
    page.screenshot(path=here / f"{browser_name}.png")
    page.close()
    browser.close()


if __name__ == "__main__":
    browser_name = sys.argv[1]
    with sync_playwright() as p:
        main(p, browser_name)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from playwright.async_api import Page, expect


async def test_get_by_escaping(page: Page) -> None:
    await page.set_content(
        """
        <label id=label for=control>Hello my
wo"rld</label><input id=control />"""
    )
    await page.eval_on_selector(
        "input",
        """input => {
        input.setAttribute('placeholder', 'hello my\\nwo"rld');
        input.setAttribute('title', 'hello my\\nwo"rld');
        input.setAttribute('alt', 'hello my\\nwo"rld');
    }""",
    )
    await expect(page.get_by_text('hello my\nwo"rld')).to_have_attribute("id", "label")
    await expect(page.get_by_text('hello       my     wo"rld')).to_have_attribute(
        "id", "label"
    )
    await expect(page.get_by_label('hello my\nwo"rld')).to_have_attribute(
        "id", "control"
    )
    await expect(page.get_by_placeholder('hello my\nwo"rld')).to_have_attribute(
        "id", "control"
    )
    await expect(page.get_by_alt_text('hello my\nwo"rld')).to_have_attribute(
        "id", "control"
    )
    await expect(page.get_by_title('hello my\nwo"rld')).to_have_attribute(
        "id", "control"
    )

    await page.set_content(
        """
        <label id=label for=control>Hello my
world</label><input id=control />"""
    )
    await page.eval_on_selector(
        "input",
        """input => {
        input.setAttribute('placeholder', 'hello my\\nworld');
        input.setAttribute('title', 'hello my\\nworld');
        input.setAttribute('alt', 'hello my\\nworld');
    }""",
    )
    await expect(page.get_by_text("hello my\nworld")).to_have_attribute("id", "label")
    await expect(page.get_by_text("hello        my    world")).to_have_attribute(
        "id", "label"
    )
    await expect(page.get_by_label("hello my\nworld")).to_have_attribute(
        "id", "control"
    )
    await expect(page.get_by_placeholder("hello my\nworld")).to_have_attribute(
        "id", "control"
    )
    await expect(page.get_by_alt_text("hello my\nworld")).to_have_attribute(
        "id", "control"
    )
    await expect(page.get_by_title("hello my\nworld")).to_have_attribute(
        "id", "control"
    )

    await page.set_content("""<div id=target title="my title">Text here</div>""")
    await expect(page.get_by_title("my title", exact=True)).to_have_count(
        1, timeout=500
    )
    await expect(page.get_by_title("my t\\itle", exact=True)).to_have_count(
        0, timeout=500
    )
    await expect(page.get_by_title("my t\\\\itle", exact=True)).to_have_count(
        0, timeout=500
    )

    await page.set_content(
        """<label for=target>foo &gt;&gt; bar</label><input id=target>"""
    )
    await page.eval_on_selector(
        "input",
        """input => {
        input.setAttribute('placeholder', 'foo >> bar');
        input.setAttribute('title', 'foo >> bar');
        input.setAttribute('alt', 'foo >> bar');
    }""",
    )
    assert await page.get_by_text("foo >> bar").text_content() == "foo >> bar"
    await expect(page.locator("label")).to_have_text("foo >> bar")
    await expect(page.get_by_text("foo >> bar")).to_have_text("foo >> bar")
    assert (
        await page.get_by_text(re.compile("foo >> bar")).text_content() == "foo >> bar"
    )
    await expect(page.get_by_label("foo >> bar")).to_have_attribute("id", "target")
    await expect(page.get_by_label(re.compile("foo >> bar"))).to_have_attribute(
        "id", "target"
    )
    await expect(page.get_by_placeholder("foo >> bar")).to_have_attribute(
        "id", "target"
    )
    await expect(page.get_by_alt_text("foo >> bar")).to_have_attribute("id", "target")
    await expect(page.get_by_title("foo >> bar")).to_have_attribute("id", "target")
    await expect(page.get_by_placeholder(re.compile("foo >> bar"))).to_have_attribute(
        "id", "target"
    )
    await expect(page.get_by_alt_text(re.compile("foo >> bar"))).to_have_attribute(
        "id", "target"
    )
    await expect(page.get_by_title(re.compile("foo >> bar"))).to_have_attribute(
        "id", "target"
    )


async def test_get_by_role_escaping(
    page: Page,
) -> None:
    await page.set_content(
        """
        <a href="https://playwright.dev">issues 123</a>
        <a href="https://playwright.dev">he llo 56</a>
        <button>Click me</button>
    """
    )
    assert await page.get_by_role("button").evaluate_all(
        "els => els.map(e => e.outerHTML)"
    ) == [
        "<button>Click me</button>",
    ]
    assert await page.get_by_role("link").evaluate_all(
        "els => els.map(e => e.outerHTML)"
    ) == [
        """<a href="https://playwright.dev">issues 123</a>""",
        """<a href="https://playwright.dev">he llo 56</a>""",
    ]

    assert await page.get_by_role("link", name="issues 123").evaluate_all(
        "els => els.map(e => e.outerHTML)"
    ) == [
        """<a href="https://playwright.dev">issues 123</a>""",
    ]
    assert await page.get_by_role("link", name="sues").evaluate_all(
        "els => els.map(e => e.outerHTML)"
    ) == [
        """<a href="https://playwright.dev">issues 123</a>""",
    ]
    assert await page.get_by_role("link", name="  he    \n  llo ").evaluate_all(
        "els => els.map(e => e.outerHTML)"
    ) == [
        """<a href="https://playwright.dev">he llo 56</a>""",
    ]
    assert (
        await page.get_by_role("button", name="issues").evaluate_all(
            "els => els.map(e => e.outerHTML)"
        )
        == []
    )

    assert (
        await page.get_by_role("link", name="sues", exact=True).evaluate_all(
            "els => els.map(e => e.outerHTML)"
        )
        == []
    )
    assert await page.get_by_role(
        "link", name="   he \n llo 56 ", exact=True
    ).evaluate_all("els => els.map(e => e.outerHTML)") == [
        """<a href="https://playwright.dev">he llo 56</a>""",
    ]

    assert await page.get_by_role("button", name="Click me", exact=True).evaluate_all(
        "els => els.map(e => e.outerHTML)"
    ) == [
        "<button>Click me</button>",
    ]
    assert (
        await page.get_by_role("button", name="Click \\me", exact=True).evaluate_all(
            "els => els.map(e => e.outerHTML)"
        )
        == []
    )
    assert (
        await page.get_by_role("button", name="Click \\\\me", exact=True).evaluate_all(
            "els => els.map(e => e.outerHTML)"
        )
        == []
    )


async def test_include_hidden_should_work(
    page: Page,
) -> None:
    await page.set_content("""<button style="display: none">Hidden</button>""")
    assert (
        await page.get_by_role("button", name="Hidden").evaluate_all(
            "els => els.map(e => e.outerHTML)"
        )
        == []
    )
    assert await page.get_by_role(
        "button", name="Hidden", include_hidden=True
    ).evaluate_all("els => els.map(e => e.outerHTML)") == [
        """<button style="display: none">Hidden</button>""",
    ]



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import subprocess
import sys
from pathlib import Path
from venv import EnvBuilder


def test_install(tmp_path: Path, browser_name: str) -> None:
    env_dir = tmp_path / "env"
    env = EnvBuilder(with_pip=True)
    env.create(env_dir=env_dir)
    context = env.ensure_directories(env_dir)
    root = Path(__file__).parent.parent.resolve()
    if sys.platform == "win32":
        wheelpath = list((root / "dist").glob("playwright*win_amd64*.whl"))[0]
    elif sys.platform == "linux":
        wheelpath = list((root / "dist").glob("playwright*manylinux1*.whl"))[0]
    elif sys.platform == "darwin":
        wheelpath = list((root / "dist").glob("playwright*macosx_*.whl"))[0]
    subprocess.check_output(
        [
            context.env_exe,
            "-m",
            "pip",
            "install",
            str(wheelpath),
        ]
    )
    environ = os.environ.copy()
    environ["PLAYWRIGHT_BROWSERS_PATH"] = str(tmp_path)
    subprocess.check_output(
        [context.env_exe, "-m", "playwright", "install", browser_name], env=environ
    )
    shutil.copyfile(root / "tests" / "assets" / "client.py", tmp_path / "main.py")
    subprocess.check_output(
        [context.env_exe, str(tmp_path / "main.py"), browser_name], env=environ
    )
    assert (tmp_path / f"{browser_name}.png").exists()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime

import pytest

from playwright.async_api import ConsoleMessage, Error, Page


async def test_should_timeout(page: Page) -> None:
    start_time = datetime.now()
    timeout = 42
    await page.wait_for_timeout(timeout)
    assert ((datetime.now() - start_time).microseconds * 1000) >= timeout / 2


async def test_should_accept_a_string(page: Page) -> None:
    watchdog = page.wait_for_function("window.__FOO === 1")
    await page.evaluate("window['__FOO'] = 1")
    await watchdog


async def test_should_work_when_resolved_right_before_execution_context_disposal(
    page: Page,
) -> None:
    await page.add_init_script("window['__RELOADED'] = true")
    await page.wait_for_function(
        """() => {
            if (!window['__RELOADED'])
                window.location.reload();
            return true;
        }"""
    )


async def test_should_poll_on_interval(page: Page) -> None:
    polling = 100
    time_delta = await page.wait_for_function(
        """() => {
            if (!window['__startTime']) {
                window['__startTime'] = Date.now();
                return false;
            }
            return Date.now() - window['__startTime'];
        }""",
        polling=polling,
    )
    assert await time_delta.json_value() >= polling


async def test_should_avoid_side_effects_after_timeout(page: Page) -> None:
    counter = 0

    async def on_console(message: ConsoleMessage) -> None:
        nonlocal counter
        counter += 1

    page.on("console", on_console)
    with pytest.raises(Error) as exc_info:
        await page.wait_for_function(
            """() => {
            window['counter'] = (window['counter'] || 0) + 1;
            console.log(window['counter']);
        }""",
            polling=1,
            timeout=1000,
        )

    saved_counter = counter
    await page.wait_for_timeout(2000)  # Give it some time to produce more logs.

    assert "Timeout 1000ms exceeded" in exc_info.value.message
    assert counter == saved_counter


async def test_should_throw_on_polling_mutation(page: Page) -> None:
    with pytest.raises(Error) as exc_info:
        await page.wait_for_function("() => true", polling="mutation")  # type: ignore
    assert "Unknown polling option: mutation" in exc_info.value.message



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import datetime
from typing import Any, AsyncGenerator, List

import pytest

from playwright.async_api import Error, Page
from tests.server import Server


@pytest.fixture(autouse=True)
async def calls(page: Page) -> List[Any]:
    calls: List[Any] = []
    await page.expose_function("stub", lambda *args: calls.append(list(args)))
    return calls


class TestRunFor:
    @pytest.fixture(autouse=True)
    async def before_each(self, page: Page) -> AsyncGenerator[None, None]:
        await page.clock.install(time=0)
        await page.clock.pause_at(1000)
        yield

    async def test_run_for_triggers_immediately_without_specified_delay(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setTimeout(window.stub)")
        await page.clock.run_for(0)
        assert len(calls) == 1

    async def test_run_for_does_not_trigger_without_sufficient_delay(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setTimeout(window.stub, 100)")
        await page.clock.run_for(10)
        assert len(calls) == 0

    async def test_run_for_triggers_after_sufficient_delay(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setTimeout(window.stub, 100)")
        await page.clock.run_for(100)
        assert len(calls) == 1

    async def test_run_for_triggers_simultaneous_timers(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate(
            "setTimeout(window.stub, 100); setTimeout(window.stub, 100)"
        )
        await page.clock.run_for(100)
        assert len(calls) == 2

    async def test_run_for_triggers_multiple_simultaneous_timers(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate(
            "setTimeout(window.stub, 100); setTimeout(window.stub, 100); setTimeout(window.stub, 99); setTimeout(window.stub, 100)"
        )
        await page.clock.run_for(100)
        assert len(calls) == 4

    async def test_run_for_waits_after_setTimeout_was_called(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setTimeout(window.stub, 150)")
        await page.clock.run_for(50)
        assert len(calls) == 0
        await page.clock.run_for(100)
        assert len(calls) == 1

    async def test_run_for_triggers_event_when_some_throw(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate(
            "setTimeout(() => { throw new Error(); }, 100); setTimeout(window.stub, 120)"
        )
        with pytest.raises(Error):
            await page.clock.run_for(120)
        assert len(calls) == 1

    async def test_run_for_creates_updated_Date_while_ticking(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.clock.set_system_time(0)
        await page.evaluate(
            "setInterval(() => { window.stub(new Date().getTime()); }, 10)"
        )
        await page.clock.run_for(100)
        assert calls == [
            [10],
            [20],
            [30],
            [40],
            [50],
            [60],
            [70],
            [80],
            [90],
            [100],
        ]

    async def test_run_for_passes_8_seconds(self, page: Page, calls: List[Any]) -> None:
        await page.evaluate("setInterval(window.stub, 4000)")
        await page.clock.run_for("08")
        assert len(calls) == 2

    async def test_run_for_passes_1_minute(self, page: Page, calls: List[Any]) -> None:
        await page.evaluate("setInterval(window.stub, 6000)")
        await page.clock.run_for("01:00")
        assert len(calls) == 10

    async def test_run_for_passes_2_hours_34_minutes_and_10_seconds(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setInterval(window.stub, 10000)")
        await page.clock.run_for("02:34:10")
        assert len(calls) == 925

    async def test_run_for_throws_for_invalid_format(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setInterval(window.stub, 10000)")
        with pytest.raises(Error):
            await page.clock.run_for("12:02:34:10")
        assert len(calls) == 0

    async def test_run_for_returns_the_current_now_value(self, page: Page) -> None:
        await page.clock.set_system_time(0)
        value = 200
        await page.clock.run_for(value)
        assert await page.evaluate("Date.now()") == value


class TestFastForward:
    @pytest.fixture(autouse=True)
    async def before_each(self, page: Page) -> AsyncGenerator[None, None]:
        await page.clock.install(time=0)
        await page.clock.pause_at(1)
        yield

    async def test_ignores_timers_which_wouldnt_be_run(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate(
            "setTimeout(() => { window.stub('should not be logged'); }, 1000)"
        )
        await page.clock.fast_forward(500)
        assert len(calls) == 0

    async def test_pushes_back_execution_time_for_skipped_timers(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setTimeout(() => { window.stub(Date.now()); }, 1000)")
        await page.clock.fast_forward(2000)
        assert calls == [[1000 + 2000]]

    async def test_supports_string_time_arguments(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate(
            "setTimeout(() => { window.stub(Date.now()); }, 100000)"
        )  # 100000 = 1:40
        await page.clock.fast_forward("01:50")
        assert calls == [[1000 + 110000]]


class TestStubTimers:
    @pytest.fixture(autouse=True)
    async def before_each(self, page: Page) -> AsyncGenerator[None, None]:
        await page.clock.install(time=0)
        await page.clock.pause_at(1)
        yield

    async def test_sets_initial_timestamp(self, page: Page) -> None:
        await page.clock.set_system_time(1.4)
        assert await page.evaluate("Date.now()") == 1400

    async def test_replaces_global_setTimeout(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setTimeout(window.stub, 1000)")
        await page.clock.run_for(1000)
        assert len(calls) == 1

    async def test_global_fake_setTimeout_should_return_id(self, page: Page) -> None:
        to = await page.evaluate("setTimeout(window.stub, 1000)")
        assert isinstance(to, int)

    async def test_replaces_global_clearTimeout(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate(
            """
            const to = setTimeout(window.stub, 1000);
            clearTimeout(to);
        """
        )
        await page.clock.run_for(1000)
        assert len(calls) == 0

    async def test_replaces_global_setInterval(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate("setInterval(window.stub, 500)")
        await page.clock.run_for(1000)
        assert len(calls) == 2

    async def test_replaces_global_clearInterval(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.evaluate(
            """
            const to = setInterval(window.stub, 500);
            clearInterval(to);
        """
        )
        await page.clock.run_for(1000)
        assert len(calls) == 0

    async def test_replaces_global_performance_now(self, page: Page) -> None:
        promise = asyncio.create_task(
            page.evaluate(
                """async () => {
            const prev = performance.now();
            await new Promise(f => setTimeout(f, 1000));
            const next = performance.now();
            return { prev, next };
        }"""
            )
        )
        await asyncio.sleep(0)  # Make sure the promise is scheduled.
        await page.clock.run_for(1000)
        assert await promise == {"prev": 1000, "next": 2000}

    async def test_fakes_Date_constructor(self, page: Page) -> None:
        now = await page.evaluate("new Date().getTime()")
        assert now == 1000


class TestStubTimersPerformance:
    async def test_replaces_global_performance_time_origin(self, page: Page) -> None:
        await page.clock.install(time=1)
        await page.clock.pause_at(2)
        promise = asyncio.create_task(
            page.evaluate(
                """async () => {
            const prev = performance.now();
            await new Promise(f => setTimeout(f, 1000));
            const next = performance.now();
            return { prev, next };
        }"""
            )
        )
        await asyncio.sleep(0)  # Make sure the promise is scheduled.
        await page.clock.run_for(1000)
        assert await page.evaluate("performance.timeOrigin") == 1000
        assert await promise == {"prev": 1000, "next": 2000}


class TestPopup:
    async def test_should_tick_after_popup(self, page: Page) -> None:
        await page.clock.install(time=0)
        now = datetime.datetime.fromisoformat("2015-09-25")
        await page.clock.pause_at(now)
        popup, _ = await asyncio.gather(
            page.wait_for_event("popup"), page.evaluate("window.open('about:blank')")
        )
        popup_time = await popup.evaluate("Date.now()")
        assert popup_time == now.timestamp() * 1000
        await page.clock.run_for(1000)
        popup_time_after = await popup.evaluate("Date.now()")
        assert popup_time_after == now.timestamp() * 1000 + 1000

    async def test_should_tick_before_popup(self, page: Page) -> None:
        await page.clock.install(time=0)
        now = datetime.datetime.fromisoformat("2015-09-25")
        await page.clock.pause_at(now)
        await page.clock.run_for(1000)
        popup, _ = await asyncio.gather(
            page.wait_for_event("popup"), page.evaluate("window.open('about:blank')")
        )
        popup_time = await popup.evaluate("Date.now()")
        assert popup_time == int(now.timestamp() * 1000 + 1000)
        assert datetime.datetime.fromtimestamp(popup_time / 1_000).year == 2015

    async def test_should_run_time_before_popup(
        self, page: Page, server: Server
    ) -> None:
        server.set_route(
            "/popup.html",
            lambda res: (
                res.setHeader("Content-Type", "text/html"),
                res.write(b"<script>window.time = Date.now()</script>"),
                res.finish(),
            ),
        )
        await page.goto(server.EMPTY_PAGE)
        # Wait for 2 second in real life to check that it is past in popup.
        await page.wait_for_timeout(2000)
        popup, _ = await asyncio.gather(
            page.wait_for_event("popup"),
            page.evaluate("window.open('{}')".format(server.PREFIX + "/popup.html")),
        )
        popup_time = await popup.evaluate("window.time")
        assert popup_time >= 2000

    async def test_should_not_run_time_before_popup_on_pause(
        self, page: Page, server: Server
    ) -> None:
        server.set_route(
            "/popup.html",
            lambda res: (
                res.setHeader("Content-Type", "text/html"),
                res.write(b"<script>window.time = Date.now()</script>"),
                res.finish(),
            ),
        )
        await page.clock.install(time=0)
        await page.clock.pause_at(1)
        await page.goto(server.EMPTY_PAGE)
        # Wait for 2 second in real life to check that it is past in popup.
        await page.wait_for_timeout(2000)
        popup, _ = await asyncio.gather(
            page.wait_for_event("popup"),
            page.evaluate("window.open('{}')".format(server.PREFIX + "/popup.html")),
        )
        popup_time = await popup.evaluate("window.time")
        assert popup_time == 1000


class TestSetFixedTime:
    async def test_does_not_fake_methods(self, page: Page) -> None:
        await page.clock.set_fixed_time(0)
        # Should not stall.
        await page.evaluate("new Promise(f => setTimeout(f, 1))")

    async def test_allows_setting_time_multiple_times(self, page: Page) -> None:
        await page.clock.set_fixed_time(0.1)
        assert await page.evaluate("Date.now()") == 100
        await page.clock.set_fixed_time(0.2)
        assert await page.evaluate("Date.now()") == 200

    async def test_fixed_time_is_not_affected_by_clock_manipulation(
        self, page: Page
    ) -> None:
        await page.clock.set_fixed_time(0.1)
        assert await page.evaluate("Date.now()") == 100
        await page.clock.fast_forward(20)
        assert await page.evaluate("Date.now()") == 100

    async def test_allows_installing_fake_timers_after_setting_time(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.clock.set_fixed_time(0.1)
        assert await page.evaluate("Date.now()") == 100
        await page.clock.set_fixed_time(0.2)
        await page.evaluate("setTimeout(() => window.stub(Date.now()))")
        await page.clock.run_for(0)
        assert calls == [[200]]


class TestWhileRunning:
    async def test_should_progress_time(self, page: Page) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.wait_for_timeout(1000)
        now = await page.evaluate("Date.now()")
        assert 1000 <= now <= 2000

    async def test_should_run_for(self, page: Page) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.run_for(10000)
        now = await page.evaluate("Date.now()")
        assert 10000 <= now <= 11000

    async def test_should_fast_forward(self, page: Page) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.fast_forward(10000)
        now = await page.evaluate("Date.now()")
        assert 10000 <= now <= 11000

    async def test_should_fast_forward_to(self, page: Page) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.fast_forward(10000)
        now = await page.evaluate("Date.now()")
        assert 10000 <= now <= 11000

    async def test_should_pause(self, page: Page) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.pause_at(1)
        await page.wait_for_timeout(1000)
        now = await page.evaluate("Date.now()")
        assert 0 <= now <= 1000

    async def test_should_pause_and_fast_forward(self, page: Page) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.pause_at(1)
        await page.clock.fast_forward(1000)
        now = await page.evaluate("Date.now()")
        assert now == 2000

    async def test_should_set_system_time_on_pause(self, page: Page) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.pause_at(1)
        now = await page.evaluate("Date.now()")
        assert now == 1000


class TestWhileOnPause:
    async def test_fast_forward_should_not_run_nested_immediate(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.pause_at(1000)
        await page.evaluate(
            """
            setTimeout(() => {
                window.stub('outer');
                setTimeout(() => window.stub('inner'), 0);
            }, 1000);
        """
        )
        await page.clock.fast_forward(1000)
        assert calls == [["outer"]]
        await page.clock.fast_forward(1)
        assert calls == [["outer"], ["inner"]]

    async def test_run_for_should_not_run_nested_immediate(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.pause_at(1000)
        await page.evaluate(
            """
            setTimeout(() => {
                window.stub('outer');
                setTimeout(() => window.stub('inner'), 0);
            }, 1000);
        """
        )
        await page.clock.run_for(1000)
        assert calls == [["outer"]]
        await page.clock.run_for(1)
        assert calls == [["outer"], ["inner"]]

    async def test_run_for_should_not_run_nested_immediate_from_microtask(
        self, page: Page, calls: List[Any]
    ) -> None:
        await page.clock.install(time=0)
        await page.goto("data:text/html,")
        await page.clock.pause_at(1000)
        await page.evaluate(
            """
            setTimeout(() => {
                window.stub('outer');
                void Promise.resolve().then(() => setTimeout(() => window.stub('inner'), 0));
            }, 1000);
        """
        )
        await page.clock.run_for(1000)
        assert calls == [["outer"]]
        await page.clock.run_for(1)
        assert calls == [["outer"], ["inner"]]



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import asyncio
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List

import pytest
from greenlet import greenlet

from playwright._impl._driver import compute_driver_executable
from playwright.sync_api import (
    Browser,
    BrowserContext,
    BrowserType,
    FrameLocator,
    Locator,
    Page,
    Playwright,
    Selectors,
    sync_playwright,
)
from tests.server import HTTPServer

from .utils import Utils
from .utils import utils as utils_object


@pytest.fixture
def utils() -> Generator[Utils, None, None]:
    yield utils_object


@pytest.fixture(scope="session")
def playwright() -> Generator[Playwright, None, None]:
    with sync_playwright() as p:
        yield p


@pytest.fixture(scope="session")
def browser_type(
    playwright: Playwright, browser_name: str
) -> Generator[BrowserType, None, None]:
    browser_type = None
    if browser_name == "chromium":
        browser_type = playwright.chromium
    elif browser_name == "firefox":
        browser_type = playwright.firefox
    elif browser_name == "webkit":
        browser_type = playwright.webkit
    assert browser_type, f"Unkown browser name '{browser_name}'"
    yield browser_type


@pytest.fixture(scope="session")
def browser(
    browser_type: BrowserType, launch_arguments: Dict
) -> Generator[Browser, None, None]:
    browser = browser_type.launch(**launch_arguments)
    yield browser
    browser.close()


@pytest.fixture
def context(browser: Browser) -> Generator[BrowserContext, None, None]:
    context = browser.new_context()
    yield context
    context.close()


@pytest.fixture
def page(context: BrowserContext) -> Generator[Page, None, None]:
    page = context.new_page()
    yield page
    page.close()


@pytest.fixture(scope="session")
def selectors(playwright: Playwright) -> Selectors:
    return playwright.selectors


@pytest.fixture(scope="session")
def sync_gather(playwright: Playwright) -> Generator[Callable, None, None]:
    def _sync_gather_impl(*actions: Callable) -> List[Any]:
        g_self = greenlet.getcurrent()
        results: Dict[Callable, Any] = {}
        exceptions: List[Exception] = []

        def action_wrapper(action: Callable) -> Callable:
            def body() -> Any:
                try:
                    results[action] = action()
                except Exception as e:
                    results[action] = e
                    exceptions.append(e)
                g_self.switch()

            return body

        async def task() -> None:
            for action in actions:
                g = greenlet(action_wrapper(action))
                g.switch()

        asyncio.create_task(task())

        while len(results) < len(actions):
            playwright._dispatcher_fiber.switch()

        if exceptions:
            raise exceptions[0]

        return list(map(lambda action: results[action], actions))

    yield _sync_gather_impl


class TraceViewerPage:
    def __init__(self, page: Page):
        self.page = page

    @property
    def actions_tree(self) -> Locator:
        return self.page.get_by_test_id("actions-tree")

    @property
    def action_titles(self) -> Locator:
        return self.page.locator(".action-title")

    @property
    def stack_frames(self) -> Locator:
        return self.page.get_by_role("list", name="Stack trace").get_by_role("listitem")

    def select_action(self, title: str, ordinal: int = 0) -> None:
        self.page.locator(".action-title", has_text=title).nth(ordinal).click()

    def select_snapshot(self, name: str) -> None:
        self.page.click(f'.snapshot-tab .tabbed-pane-tab-label:has-text("{name}")')

    def snapshot_frame(
        self, action_name: str, ordinal: int = 0, has_subframe: bool = False
    ) -> FrameLocator:
        self.select_action(action_name, ordinal)
        expected_frames = 4 if has_subframe else 3
        while len(self.page.frames) < expected_frames:
            self.page.wait_for_event("frameattached")
        return self.page.frame_locator("iframe.snapshot-visible[name=snapshot]")

    def show_source_tab(self) -> None:
        self.page.click("text='Source'")

    def expand_action(self, title: str, ordinal: int = 0) -> None:
        self.actions_tree.locator(".tree-view-entry", has_text=title).nth(
            ordinal
        ).locator(".codicon-chevron-right").click()


@pytest.fixture
def show_trace_viewer(browser: Browser) -> Generator[Callable, None, None]:
    """Fixture that provides a function to show trace viewer for a trace file."""

    @contextmanager
    def _show_trace_viewer(
        trace_path: Path,
    ) -> Generator[TraceViewerPage, None, None]:
        trace_viewer_path = (
            Path(compute_driver_executable()[0]) / "../package/lib/vite/traceViewer"
        ).resolve()

        server = HTTPServer()
        server.start(trace_viewer_path)
        server.set_route("/trace.zip", lambda request: request.serve_file(trace_path))

        page = browser.new_page()

        try:
            page.goto(f"{server.PREFIX}/index.html?trace={server.PREFIX}/trace.zip")
            yield TraceViewerPage(page)
        finally:
            page.close()
            server.stop()

    yield _show_trace_viewer



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path
from typing import Any

from playwright.sync_api import Page


def test_expect_file_chooser(page: Page) -> None:
    page.set_content("<input type=file></input>")
    with page.expect_file_chooser() as fc_info:
        page.click('input[type="file"]')
    fc = fc_info.value
    fc.set_files(
        {"name": "test.txt", "mimeType": "text/plain", "buffer": b"Hello World"}
    )


def test_set_input_files_should_preserve_last_modified_timestamp(
    page: Page,
    assetdir: Path,
) -> None:
    page.set_content("<input type=file multiple=true/>")
    input = page.locator("input")
    files: Any = ["file-to-upload.txt", "file-to-upload-2.txt"]
    input.set_input_files([assetdir / file for file in files])
    assert input.evaluate("input => [...input.files].map(f => f.name)") == files
    timestamps = input.evaluate("input => [...input.files].map(f => f.lastModified)")
    expected_timestamps = [os.path.getmtime(assetdir / file) * 1000 for file in files]

    # On Linux browser sometimes reduces the timestamp by 1ms: 1696272058110.0715  -> 1696272058109 or even
    # rounds it to seconds in WebKit: 1696272058110 -> 1696272058000.
    for i in range(len(timestamps)):
        assert abs(timestamps[i] - expected_timestamps[i]) < 1000
